// Copyright (C) 2010 Johan Hake
// Licensed under the GNU LGPL Version 3.0
//
// First added:  2010-01-12
// Last changed: 2012-04-24

#include <algorithm>

#include "GillespieStepper.h"
#include "MarkovModel.h"

using namespace gillstep;

//-----------------------------------------------------------------------------
GillespieStepper::GillespieStepper():
  collect_values(false),
  collect_states(false),
  _rand_generator(), _dimless_time(0.0),
  _time(0.0), _use_deterministic(false), 
  _deterministic_number(0.5), 
  _max_fractional_rate_change(0.1), _mms(), 
  _transition_to_mm(), _mm_to_transition(),
  _offsets(), _rates(), _cumsum_rates(),
  _collected_values(),
  _collected_value_times(),
  _collected_states(),
  _collected_state_times(),
  _latest_stepped_model(0)
{}
//-----------------------------------------------------------------------------
//GillespieStepper::GillespieStepper(const uint32 seed):_rand_generator(seed),  
//				   _dimless_time(0.0), _time(0.0), 
//				   _use_deterministic(false), 
//				   _deterministic_number(0.5), _mms(0), 
//				   _transition_to_mm(0), _mm_to_transition(),
//				   _offsets(0), _rates(0), _cumsum_rates(0)
//{}
////-----------------------------------------------------------------------------
//GillespieStepper::GillespieStepper(uint32 *const big_seed, 
//				   uint32  const seed_length):
//                                   _rand_generator(big_seed, seed_length), 
//				   _dimless_time(0.0),  _time(0.0), 
//				   _use_deterministic(false), 
//				   _deterministic_number(0.5), _mms(0),
//				   _transition_to_mm(0), _mm_to_transition(),
//				   _offsets(0), _rates(0), _cumsum_rates(0)
//{}
////-----------------------------------------------------------------------------
//GillespieStepper::GillespieStepper(MTRandomGenerator& rand):_rand_generator(rand), 
//				   _dimless_time(0.0), _time(0.0),
//				   _use_deterministic(false), 
//				   _deterministic_number(0.5), _mms(0), 
//				   _transition_to_mm(0), _mm_to_transition(),
//				   _offsets(0), _rates(0), _cumsum_rates(0)
//
//{}
//-----------------------------------------------------------------------------
GillespieStepper::~GillespieStepper()
{}
//-----------------------------------------------------------------------------
void GillespieStepper::_add_markov_model(MarkovModel* mm)
{
  _mms.push_back(mm);
}
//-----------------------------------------------------------------------------
void GillespieStepper::_remove_markov_model(MarkovModel* mm)
{
  std::vector<MarkovModel*>::iterator it = std::find(_mms.begin(), _mms.end(), mm);
  if (it != _mms.end())
    _mms.erase(it);
}
//-----------------------------------------------------------------------------
void GillespieStepper::reset(double time)
{
  // Local variables
  uint i;
  std::vector<MarkovModel*>::iterator it;

  // Reset gillstep time
  _time = time;

  // Clear vectors that will be populated and reserve an estimate of what we need
  _rates.clear(); 
  _rates.reserve(2*num_markov_models());
  _offsets.clear(); 
  _offsets.reserve(2*num_markov_models());
  _transition_to_mm.clear(); 
  _transition_to_mm.reserve(2*num_markov_models());
  _mm_to_transition.clear();
  _cumsum_rates.clear();
  _collected_states.clear();
  _collected_state_times.clear();
  _collected_values.clear();
  _collected_value_times.clear();
 
  // Gather transition information, reset the markov models and send their values
  uint max_transitions = 0;
  for (it = _mms.begin(); it != _mms.end(); it++){

    // Store the starting transition for each Markov model
    _mm_to_transition[*it] = max_transitions;
    max_transitions += (*it)->max_transitions();

    // Do not reset sub models. They are reseted by the parent model
    if (!(*it)->is_submodel()){
      (*it)->reset();
      (*it)->send();
    }

    // Initiate the collected value maps
    if (collect_values || collect_states){
      
      // Local vector which will be passed by value
      std::vector<double> values;
      std::vector<uint> values_uint;
      
      // Attach new vectors for each Markov model
      _collected_states[*it] = values_uint;
      _collected_state_times[*it] = values;
      _collected_values[*it] = values;
      _collected_value_times[*it] = values;

      // Reserve memory for collected values
      if (collect_states){
	_collected_states[*it].reserve(CHUNK);
	_collected_state_times[*it].reserve(CHUNK);
	_collected_states[*it].push_back((*it)->state());
	_collected_state_times[*it].push_back(_time);
      }

      if (collect_values){
	_collected_values[*it].reserve(CHUNK);
	_collected_value_times[*it].reserve(CHUNK);
	_collected_values[*it].push_back((*it)->value());
	_collected_value_times[*it].push_back(_time);
      }
    }
  }

  // Walk through all rates and collect their current value
  // and populate the Markov model map and offset vector
  for (it = _mms.begin(); it != _mms.end(); it++)
  {
    for(i = 0; i <(*it)->max_transitions(); i++)
    {
      _rates.push_back((*it)->evaluate_rate(i));
      _offsets.push_back(i);
      _transition_to_mm.push_back((*it));
    }
  }

  // Init the cumsum vector
  _cumsum_rates.resize(_rates.size(), 0.0);
  cumsum(_cumsum_rates, _rates);
  
  // Calculate a new dimensionless time step
  new_dimless_exit_time();
}
//-----------------------------------------------------------------------------
void GillespieStepper::update_rates()
{
  // Local variables
  uint i, j = 0;
  std::vector<MarkovModel*>::iterator it;

  // Walk through all rates and collect their current value
  // and populate the Markov model map and offset vector
  for (it = _mms.begin(); it != _mms.end(); it++)
    for(i = 0; i <(*it)->max_transitions(); i++)    
    {
      _rates[j++] = (*it)->evaluate_rate(i);
      if (_rates[j]<0)
	throw std::runtime_error("One of \"" + (*it)->name() + "\" rates evaluated to negative values.");
    }

  // Init the cumsum vector
  assert(_rates.size() == _cumsum_rates.size());
  cumsum(_cumsum_rates, _rates);
}
//-----------------------------------------------------------------------------
void GillespieStepper::update_rates(double present_substep, double next_substep)
{
  // Local variables
  uint i, j = 0;
  std::vector<MarkovModel*>::iterator it;

  // Walk through all rates and collect their current value
  // and populate the Markov model map and offset vector
  for (it = _mms.begin(); it != _mms.end(); it++)
    for(i = 0; i <(*it)->max_transitions(); i++)
    {
      _rates[j++] = (*it)->evaluate_substep_rate(i, present_substep, next_substep);
      if (_rates[j]<0)
	throw std::runtime_error("One of \"" + (*it)->name() + "\" rates evaluated to negative values.");
    }
  // Init the cumsum vector
  assert(_rates.size() == _cumsum_rates.size());
  cumsum(_cumsum_rates, _rates);
}
//-----------------------------------------------------------------------------
void GillespieStepper::step_dt(double dt, bool substep)
{

  // Step until dt or next channel transition
  dt = step_dt_unless_transition(dt, substep);
  
  while (dt > 0.0)
  {
    // Send the value of the stepped Markov model
    assert(send_stepped_value());
    dt = step_dt_unless_transition(dt, substep);
  }
}
//-----------------------------------------------------------------------------
double GillespieStepper::substep_size(double present_substep)
{
  std::vector<MarkovModel*>::iterator it;

  // Walk through Markov models and collect the smallest substep needed to be taken
  double next_substep = 1.0;
  for (it = _mms.begin(); it != _mms.end(); it++)
    next_substep = std::min(next_substep, (*it)->substep_size(present_substep));

  return next_substep;
}
//-----------------------------------------------------------------------------
double GillespieStepper::step_dt_unless_transition(double dt, bool substep)
{
  assert(dt > 0.0);
  
  // If no MarkovModel is not registered just step the time
  if (_mms.size() == 0) {
    _time += dt;
    return 0.0;
  }

  // Local timestep
  double dt_rest = dt;

  // Channel transition flag
  bool channel_trans = false;

  // Time to next transition
  double next_trans;

  // Check if we need to take a substep
  if (substep && this->substep_size(0.0) < 1.0) {

    // Calculate the size of the substep
    double present_substep = 0.0;
    double next_substep = this->substep_size(present_substep);
    double substep_dt = next_substep*dt;

    // std::cout << std::endl << "Present substep:" << present_substep << std::endl;
    // std::cout << "next_substep:" << next_substep << std::endl;
    // std::cout << "substep_dt:" << substep_dt << std::endl;
    
    // Update the rates using the substep
    update_rates(present_substep, next_substep);

    // Time to next transition
    next_trans = next_transition();

    // std::cout << "\nBefore loop:" << std::endl;
    // std::cout << "next trans:" << next_trans << std::endl;
    // std::cout << "substep_dt:" << substep_dt << std::endl;
    
    // While not channel trans or we have run out of our dt
    while (!(channel_trans || is_zero(dt_rest)))
    {
      
      // std::cout << "\nOuter loop:" << std::endl;
      // std::cout << "is_zero(dt_rest):" << is_zero(dt_rest) << std::endl;
      // std::cout << "next_substep:" << next_substep << std::endl;
      // std::cout << "dt_rest:" << dt_rest << std::endl;
      // std::cout << "next trans:" << next_trans << std::endl;
      // std::cout << "substep_dt:" << substep_dt << std::endl;
    
      // Inner substep loop
      while (next_trans > substep_dt && !is_zero(dt_rest))
      {

	// Reduce the time left
	dt_rest -= substep_dt;

	// Update gillstep time
	_time += substep_dt;
	
	// Reduce the time to next transition
	_dimless_time -= substep_dt*sum_rates();
	
	// Update old substep
	present_substep = next_substep;

	// Get the next substep given in parts of dt
	next_substep = this->substep_size(present_substep);

	// Update the rates using the substep
	update_rates(present_substep, next_substep);

	// Time to next transition
	next_trans = next_transition();

	// Get the next small timestep
	substep_dt = (next_substep-present_substep)*dt;

	// std::cout << "next_substep:" << next_substep << std::endl;
	// std::cout << "dt_rest:" << dt_rest << std::endl;
	// std::cout << "next trans:" << next_trans << std::endl;
	// std::cout << "substep_dt:" << substep_dt << std::endl << std::endl;;
      }

      // If we have not run out of the time step or next transition 
      // is smaller than next substep
      if (!is_zero(dt_rest)) {

	// Reduce the time left
	dt_rest -= next_trans;

	// Update gillstep time
	_time += next_trans;

	// Update the present substep to the time of transition
	present_substep += next_trans/dt;

	// Step Markov model
	// We do not know the next substep so pass the present one
	channel_trans = step_markov_model(present_substep, present_substep);
    
	// std::cout << "\nTransition: " << _latest_stepped_model->name() << std::endl;
	// std::cout << "channel transition: " << channel_trans <<std::endl;
	// std::cout << "dt_rest: " << dt_rest <<std::endl;
	
	// Calculate new dimension less time
	new_dimless_exit_time();
	
	// If no channel transition we update the variables for further timestepping
	if (!channel_trans){
	  
	  // Get the next substep given in parts of dt
	  next_substep = this->substep_size(present_substep);

	  // Update the rates using the substep at the transition
	  update_rates(present_substep, next_substep);

	  // Calculate new transition time
	  next_trans = next_transition();

	  // Get the next small timestep
	  substep_dt = (next_substep-present_substep)*dt;
	  
	  // std::cout << "next_substep:" << next_substep << std::endl;
	  // std::cout << "dt_rest:" << dt_rest << std::endl;
	  // std::cout << "next trans:" << next_trans << std::endl;
	  // std::cout << "substep_dt:" << substep_dt << std::endl;
    
	}
      }
    }
    
    // If channel transition return whats left of the time step and update
    // the continuous values stored in the Markov models
    if (channel_trans){
      //std::cout << "Channel transition!" << std::endl;
      std::vector<MarkovModel*>::iterator it;
      for (it = _mms.begin(); it != _mms.end(); it++)
	(*it)->update_variables((dt-dt_rest)/dt);
      
      // Send the stepped value
      send_stepped_value();
	
      return dt_rest;
     
    }
  }
  // No substep
  else
  {

    // Update the rates
    update_rates();

    // Time to next transition
    next_trans = next_transition();

    // While not channel trans or we have run out of our time step
    while (!channel_trans && next_trans < dt_rest)
    {
      // Reduce the time left
      dt_rest -= next_trans;

      // Update gillstep time
      _time += next_trans;

      // Step Markov model
      channel_trans = step_markov_model();
    
      // Calculate new dimension less time
      new_dimless_exit_time();
    
      // Calculate new transition time
      next_trans = next_transition();
    }

    // If channel transition return whats left of the time step
    if (channel_trans) {
      std::vector<MarkovModel*>::iterator it;
      for (it = _mms.begin(); it != _mms.end(); it++)
	(*it)->update_variables((dt-dt_rest)/dt);
      
      // Send the stepped value
      send_stepped_value();
	
      return dt_rest;
    }
  
  }

  // If no channel transition we have run out of time and we update
  // _dimless_time left and return zero offset
  _dimless_time -= dt_rest*sum_rates();

  // Update gillstep time
  _time += dt_rest;

  return 0.0;
}
//-----------------------------------------------------------------------------
double GillespieStepper::step_until_next_transition(bool send_value)
{
  double dt = 0.0;
  bool channel_trans = false;

  // While not channel trans
  while (!channel_trans)
  {
    // Add the time to next transition
    dt +=  next_transition();

    // Step Markov model
    channel_trans = step_markov_model();
    
    // Calculate new dimension less time
    new_dimless_exit_time();

  }

  // Send value if required
  if (send_value)
    assert(send_stepped_value());

  // Update gillstep time
  _time += dt;
  
  // Return the accumulated time
  return dt;
}
//-----------------------------------------------------------------------------
bool GillespieStepper::step_markov_model(double present_substep, double next_substep)
{
  assert(present_substep>=0);
  assert(next_substep>=present_substep);

  uint parent_transition = 0;
  
  // Pick a random transition
  uint transition = random_transition(_cumsum_rates);

  // Get the Markov model for this transition
  MarkovModel* mm = _transition_to_mm[transition];

  //std::cout << "stepping: " << mm->name() << std::cout;

  // Make the transition and record whether it is a channel transition
  bool channel_transition = mm->change_state(_offsets[transition]);

  // Get the index where the Markov model starts in the vector 
  uint mm_ind = transition - _offsets[transition];
  
  // Update rates
  if (next_substep > 0.0)
    for(uint i = 0; i <mm->max_transitions(); i++)
      _rates[mm_ind+i] = mm->evaluate_substep_rate(i, present_substep, next_substep);
  else
    for(uint i = 0; i <mm->max_transitions(); i++)
      _rates[mm_ind+i] = mm->evaluate_rate(i);

  /*
  // Check if any submodels recieve the parent 
  bool send_to_submodels = false;
  uint i = 0;
  while (!send_to_submodels && i < mm->num_submodels())
    send_to_submodels = mm->submodel(i++).receive_parent();

  // If so update the rates of the submodels
  // FIXME: Is this really nessesary!?
  // It clutter the whole interface...
  if (send_to_submodels)
  {
    // Send value
    mm->send();
    
    //std::cout << "Sending: " << mm->value() << " to submodels of " << mm->name() << std::endl;

    // Iterate over the submodels and update the rates
    for (uint j = 0; j < mm->num_submodels(); j++)
    {
      
      // Get the ith submodels
      MarkovModel& submodel = mm->submodel(j);
      
      if (submodel.receive_parent())
      {
	uint submodel_transition = _mm_to_transition[&submodel];

	// Update rates
	if (next_substep > 0.0)
	  for(i = 0; i < submodel.max_transitions(); i++)
	    _rates[submodel_transition+i] = submodel.evaluate_substep_rate(\
					     i, present_substep, next_substep);
	else
	  for(i = 0; i < submodel.max_transitions(); i++)
	    _rates[submodel_transition+i] = submodel.evaluate_rate(i);

	// Update the transition so we update the rates correctly
	transition = submodel_transition < transition ? submodel_transition : transition;
      }
    }
  }
  */

  // If sub Markov model update parents rate
  // A sub markov model does not have to alter the parents rate but we do not know.
  if (mm->is_submodel()){

    // Get parent and parent transitions
    MarkovModel& parent = mm->get_parent();
    parent_transition = _mm_to_transition[&parent];
    
    // Update rates
    if (next_substep > 0.0)
      for(uint i = 0; i <parent.max_transitions(); i++)
	_rates[parent_transition+i] = parent.evaluate_substep_rate(i, present_substep, 
								   next_substep);
    else
      for(uint i = 0; i <parent.max_transitions(); i++)
	_rates[parent_transition+i] = parent.evaluate_rate(i);

    // Update value for transition if parent transition has a smaller number
    // This will only be used to update changed 
    transition = parent_transition < transition ? parent_transition : transition;
  }

  // If transition is the zeroeth we just bump it one so our cum_sum update will work
  transition += transition == 0 ? 1 : 0;

  // Update the cumsum vector
  // FIXME: Updating all cumsums here. Prob. not strictly nessesary but 
  // FIXME: doing for safety.
  cumsum(_cumsum_rates, _rates);
  //for (uint i = transition; i < _rates.size(); i++)
  //  _cumsum_rates[i] = _cumsum_rates[i-1] + _rates[i];
  
  // Set the latest update Markov model
  _latest_stepped_model = mm;

  // Collect values
  if (channel_transition && collect_values){
    //std::cout << "Collected value " << mm->value() << " from " << mm->name() << std::endl;
    _collected_values[mm].push_back(mm->value());
    _collected_value_times[mm].push_back(_time);
  }
  
  // Collect states
  if (collect_states){
    _collected_states[mm].push_back(mm->state());
    _collected_state_times[mm].push_back(_time);
  }

  // Return if the stepped Markov model resulted in a channel transition
  return channel_transition;

}
//-----------------------------------------------------------------------------
bool GillespieStepper::send_stepped_value() const
{
  /// Check that a model has been stepped
  if (_latest_stepped_model)
  {
    _latest_stepped_model->send();
    return true;
  }

  /// Return false if not
  return false;
}
//-----------------------------------------------------------------------------
gillstep::uint GillespieStepper::random_transition(
				  const std::vector<double>& cumsum) const
{

  // A weighted random number 
  double weighted_rand_number = cumsum[cumsum.size()-1]*_rand();

  // An iterator starting at the value in cumsum that is just 
  // below weighted_rand_number
  std::vector<double>::const_iterator low = std::lower_bound (cumsum.begin(), 
							      cumsum.end(), 
							      weighted_rand_number);
  return low - cumsum.begin();
}
//-----------------------------------------------------------------------------
void GillespieStepper::set_deterministic(bool deterministic, double number)
{
  // Set the deterministic flag
  _use_deterministic = deterministic;
  _deterministic_number = number;
}
//-----------------------------------------------------------------------------
const std::vector<gillstep::uint>& GillespieStepper::get_states(MarkovModel* model)
{ 
  std::map<gillstep::MarkovModel*, std::vector<gillstep::uint> >::const_iterator it = \
    _collected_states.find(model);
  if (it == _collected_states.end())
    throw std::runtime_error("\"" + model->name() + "\" is not a MarkovModel "
			     "with registered states");
  return it->second;
}
//-----------------------------------------------------------------------------
const std::vector<double>& GillespieStepper::get_state_times(MarkovModel* model)
{ 
  std::map<gillstep::MarkovModel*, std::vector<double> >::const_iterator it = \
    _collected_state_times.find(model);
  if (it == _collected_state_times.end())
    throw std::runtime_error("\"" + model->name() + "\" is not a MarkovModel "
			     "with registered states");
  return it->second;
}
//-----------------------------------------------------------------------------
const std::vector<double>& GillespieStepper::get_values(MarkovModel* model)
{ 
  std::map<gillstep::MarkovModel*, std::vector<double> >::const_iterator it = \
    _collected_values.find(model);
  if (it == _collected_values.end())
    throw std::runtime_error("\"" + model->name() + "\" is not a MarkovModel "
			     "with registered values");
  return it->second;
}
//-----------------------------------------------------------------------------
const std::vector<double>& GillespieStepper::get_value_times(MarkovModel* model)
{ 
  std::map<gillstep::MarkovModel*, std::vector<double> >::const_iterator it = \
    _collected_value_times.find(model);
  if (it == _collected_value_times.end())
    throw std::runtime_error("\"" + model->name() + "\" is not a MarkovModel "
			     "with registered values");
  return it->second;
}
//-----------------------------------------------------------------------------
