// Copyright (C) 2010 Johan Hake
// Licensed under the GNU LGPL Version 3.0
//
// First added:  2010-01-12
// Last changed: 2011-03-01

#ifndef __GILLESPIESTEPPER_H
#define __GILLESPIESTEPPER_H

#include <cassert>
#include <vector>
#include <map>
#include <cmath>
#include <numeric>

#include "MersenneTwister.h"
#include "common.h"

namespace gillstep
{

  /// Forward declaration
  class MarkovModel;

  /// The class to hand the evolvments of the discrete Markov models
  class GillespieStepper
  {
  public:
    
    /// Create a time stepper using a random seed
    GillespieStepper();

    /// Create a time stepper using handed seed
    //GillespieStepper(const uint32 seed);
    //
    ///// Create a time stepper using handed seed vector
    //GillespieStepper(uint32 *const big_seed, 
    //		     uint32  const seed_length=MTRandomGenerator::N);
    //
    ///// Create a time stepper using passed random generator
    //GillespieStepper(MTRandomGenerator& rand);

    /// Create a time stepper using a random seed
    ~GillespieStepper();

    /// Reset the GillespieStepper based on the registered Markov models
    void reset(double time=0.0);

    /// Update rates and cumsum rates
    void update_rates();

    /// Update rates and cumsum rates using a substep
    void update_rates(double present_substep, double next_substep);

    /// Step the registered Markov models with the passed time step, and 
    /// changed values of stepped Markov models will be sent
    void step_dt(double dt, bool substep=false);

    /// Return the smallest substep needed to step the model with the registered 
    /// maximal fractional reate change
    double substep_size(double present_substep);

    /// Step the registered Markov models with the passed time step
    /// returns offset of possible channel transition. Any changed Markov
    /// model value will be sent. If substep is true the Markov models 
    /// will be stepped using a maximal fractional change in rate.
    /// FIXME: Document this in a better way
    double step_dt_unless_transition(double dt, bool substep=false);

    /// Step the registered Markov models until next channel transition
    /// returns the time it took, the new value of the stepped Markov model
    /// will be sent if send_value is true
    double step_until_next_transition(bool send_value=false);

    /// Make a random transition of a registered Markov model, return true if the 
    /// transition was a channel transition. Pass present and next_substep
    /// to be able to update the new rate functions
    bool step_markov_model(double present_substep, double next_substep);

    /// Make a random transition of a registered Markov model, return true if the 
    /// transition was a channel transition
    inline bool step_markov_model(){return step_markov_model(0.0, 0.0);};

    /// Return a random transition given a vector of cumsum of rates
    uint random_transition(const std::vector<double>& cumsum) const; 

    /// Use deterministic instead of random number
    void set_deterministic(bool deterministic, double number = 0.5);

    /// Send value of stepped Markov model, if no stepped Markov model exist, 
    /// return false, true otherwise
    /// FIXME: Make this private?
    bool send_stepped_value() const;

    /// Reset time
    inline void reset_time(double time=0.0){_time = time;}

    /// Return the local time of the GillespieStepper
    inline double time() const {return _time;}

    /// Set max fractional change in rate when externals are changing
    inline void set_max_fractional_rate_change(double value);

    /// Get max fractional change in rate when externals are changing
    inline double get_max_fractional_rate_change() const;

    /// Return the sum of rates 
    /// (if rates has changed due to changed Markov model parameters or
    /// external values, update_rates need to be called to get correct value)
    inline double sum_rates() const {return _cumsum_rates[_cumsum_rates.size()-1];}

    /// Return the time to the next transition
    inline double next_transition() const;

    /// Return the number of registered Markov models
    inline uint num_markov_models() const {return _mms.size();}

    /// Return the sum of the passed vector
    inline double sum(const std::vector<double>& vec) const;
    
    /// Return the cumsum of passed values
    inline void cumsum(std::vector<double>& cumsum,
		       const std::vector<double>& vec) const;
    
    /// Compute a new dimensionless exit time
    inline void new_dimless_exit_time(){_dimless_time = -log(_rand());};

    /// Return the state vector for the passed Markov model
    const std::vector<uint>& get_states(MarkovModel* model);

    /// Return the time vector correspondning to the state vector for the 
    /// passed Markov model
    const std::vector<double>& get_state_times(MarkovModel* model);

    /// Return the state vector for the passed Markov model
    const std::vector<double>& get_values(MarkovModel* model);

    /// Return the time vector correspondning to the state vector for the 
    /// passed Markov model
    const std::vector<double>& get_value_times(MarkovModel* model);

    /// Public attributes 
    
    // Collect value flag
    bool collect_values;

    // Collect state flag
    bool collect_states;

  private:

    /// Friends
    friend class MarkovModel;

    /// Register Markov models
    void _add_markov_model(MarkovModel* mm);

    /// Remove a registered Markov model
    void _remove_markov_model(MarkovModel* mm);

    /// Return a homogenously distributed random number between 0-1
    inline double _rand() const
    {
      return _use_deterministic ? _deterministic_number : _rand_generator.randDblExc();
    }

    /// Data members

    // The Mersenne Twister quasi random number generator
    mutable MTRandomGenerator _rand_generator;
    
    // Dimension less time
    double _dimless_time;

    // Gillstepper Time
    double _time;

    // Values for using deterministic instead of random number
    bool _use_deterministic;
    double _deterministic_number;

    // Value for max fractional rate_change
    double _max_fractional_rate_change;

    // The vector of MarkoModels
    std::vector<MarkovModel*> _mms;

    // A vector of pointers to Markov models used to map transitions to Markov models
    std::vector<MarkovModel*> _transition_to_mm;

    // A map with MarkovModel* to transitions 
    std::map<MarkovModel*, uint> _mm_to_transition;

    // Vector of transition offsets
    std::vector<uint> _offsets;
 
    // The rate vector
    std::vector<double> _rates;

    // The cumsum of rate vector
    // FIXME: Should this be a scratch vector instead?
    std::vector<double> _cumsum_rates;

    // Collected values
    std::map<MarkovModel*, std::vector<double> > _collected_values;

    // Collected value times
    std::map<MarkovModel*, std::vector<double> > _collected_value_times;

    // Collected states
    std::map<MarkovModel*, std::vector<uint> > _collected_states;

    // Collected states times
    std::map<MarkovModel*, std::vector<double> > _collected_state_times;

    /// A pointer to the latest stepped Markov model
    MarkovModel* _latest_stepped_model;

  };

  // Define inline methods
  inline void GillespieStepper::set_max_fractional_rate_change(double value)
  {
    assert(value > 0.0);
    _max_fractional_rate_change = value;
  }

  inline double GillespieStepper::get_max_fractional_rate_change() const
  {
    return _max_fractional_rate_change;
  }

  inline double GillespieStepper::next_transition() const
  {
    return (_mms.size() == 0 || _cumsum_rates[_cumsum_rates.size()-1] == 0.0) 
      ? GILLSTEP_HUGE : _dimless_time/_cumsum_rates[_cumsum_rates.size()-1];
  }

  inline double GillespieStepper::sum(const std::vector<double>& vec) const
  {
    return std::accumulate(vec.begin(), vec.end(), 0.0);
  }

  inline void GillespieStepper::cumsum(std::vector<double>& cumsum, 
				       const std::vector<double>& vec) const
  {
    std::partial_sum(vec.begin(), vec.end(), cumsum.begin());
  }

}

#endif
