// Copyright (C) 2010 Johan Hake
// Licensed under the GNU LGPL Version 3.0
//
// First added:  2010-03-15
// Last changed: 2012-05-02

#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/lu.hpp>
#include <vector>
#include <numeric>
#include <algorithm>

#include "MarkovModel.h"
#include "GillespieStepper.h"

/// uBLAS namespace 
namespace ublas = boost::numeric::ublas;

using namespace gillstep;

//-----------------------------------------------------------------------------
MarkovModel::MarkovModel(const std::string& name, GillespieStepper& gillstepper) :
  init_state(-1), fix_state(false), _retabulate_derivatives(true), _state(0), 
  _old_value(0.0), _present_variables(0), _variables_0(0), _variables_1(0), 
  _discrete_variables(0), _variable_map(), _discrete_variable_map(), 
  _value_receivers(), _submodels(0), _gillstepper(&gillstepper), _parent(0), 
  _receive_parent(false), _name(name)
{
  // Register the MarkovModel to the gillespie stepper
  _gillstepper->_add_markov_model(this);
}
//-----------------------------------------------------------------------------
MarkovModel::~MarkovModel()
{
  // Remove itself from the gillespie stepper
  _gillstepper->_remove_markov_model(this);
}
//-----------------------------------------------------------------------------
void MarkovModel::reset()
{
  // Local variables
  double sum_rates, rate;
  uint m, n;
  
  // Update variables
  for (n=0; n < _variables_0.size(); n++)
    _variables_0[n] = _variables_1[n];

  // Reset any sub models
  for (std::vector<MarkovModel*>::iterator it = _submodels.begin();
       it != _submodels.end(); it++)
    (*it)->reset();

  // Check we got a gillstepper
  if (_gillstepper == 0)
    throw std::runtime_error("No GillespiStepper registered for this MarkovModel");
  
  // Check for user specified init state value
  if (init_state > -1)
  {
    
    if (init_state >= static_cast<int>(num_states()))
      throw std::runtime_error("The value given as \"init_state\" is larger "
			       "than the number of states.");
    _state = init_state;
    
    // Update old_value
    _old_value = value();

    return;
  }

  const uint N = num_states();

  // Instantiate linear system
  ublas::vector<double> b(N, 0.0);
  ublas::matrix<double> A(N, N);

  b[N-1] = 1.0;
  
  // Loop over the states and fill matrix with rates
  for (n = 0; n < N; n++)
  {
    sum_rates = 0.0;
    for (m = 0; m < N; m++)
    {
      // Evaluate rate from state n to state m
      rate = evaluate_rate(n,m);
      A(m, n) = rate;
      sum_rates += rate;
    }
    
    // Diagonal should be filled with minus sum of rates
    A(n, n) = -sum_rates;
    
  }

  // Fill the last row with ones
  for (n = 0; n < N; n++)
    A(N-1, n) = 1.0;
  
  // Last row should all be ones
  // Create permutation matrix
  ublas::permutation_matrix<std::size_t> pmatrix(N);

  // Factorise (with pivoting)
  uint singular = ublas::lu_factorize(A, pmatrix);
  assert( singular == 0);

  // Back substitute
  ublas::lu_substitute(A, pmatrix, b);

  // Calculate the cumsum of of b
  std::vector<double> cumsum(N, 0.0);
  std::partial_sum (b.begin(), b.end(), cumsum.begin());
  
  // Return a random state based on the cumsum
  _state = _gillstepper->random_transition(cumsum);
  
  // Update old_value
  _old_value = value();

}
//-----------------------------------------------------------------------------
ValueReceiver MarkovModel::receive(std::string name)
{
  std::map<std::string, ValueReceiver>::const_iterator it = \
    _value_receivers.find(name);
  if (it == _value_receivers.end())
    throw std::runtime_error("\"" + name + "\" is not a variable this "
			     "Markov model is receiving");
  return it->second;
}
//-----------------------------------------------------------------------------
double MarkovModel::get_variable(std::string name) const 
{
  // First search variables
  std::map<std::string, uint>::const_iterator it =	\
    _variable_map.find(name);
  if (it != _variable_map.end())
    return _variables_1[it->second];
  
  // Second search discrete variables
  it = _discrete_variable_map.find(name);
  if (it == _discrete_variable_map.end())
    throw std::runtime_error("\"" + name + "\" is not a variable this "
			     "Markov model is receiving");
  return _discrete_variables[it->second];

}
//-----------------------------------------------------------------------------
void MarkovModel::set_variable(std::string name, double value)
{
  // First search variables
  std::map<std::string, uint>::const_iterator it =	\
    _variable_map.find(name);
  if (it != _variable_map.end())
  {
    _retabulate_derivatives = true;
    _variables_0[it->second] = _variables_1[0];
    _variables_1[it->second] = value;
  }
  
  // Second search discrete variables
  it = _discrete_variable_map.find(name);
  if (it == _discrete_variable_map.end())
    throw std::runtime_error("\"" + name + "\" is not a variable this "
			     "Markov model is receiving");
  _retabulate_derivatives = true;
  _discrete_variables[it->second] = value;

}
//-----------------------------------------------------------------------------
bool MarkovModel::can_receive(std::string name) const
{
  std::map<std::string, ValueReceiver>::const_iterator it =	\
    _value_receivers.find(name);
  return it != _value_receivers.end();
}
//-----------------------------------------------------------------------------
void MarkovModel::update_variables(double substep)
{
  assert(substep>=0 && substep<=1);

  // Iterate over all values and scale them according to the substep
  for (uint var = 0; var < _variables_1.size(); var++)
    _variables_1[var] = _variables_0[var] + substep*(_variables_1[var] -_variables_0[var]);
}
//-----------------------------------------------------------------------------
std::vector<std::string> MarkovModel::variable_names() const
{
  std::vector<std::string> names;
  std::map<std::string, ValueReceiver>::const_iterator it;
  for (it = _value_receivers.begin(); it != _value_receivers.end(); it++)
    names.push_back(it->first);
  return names;
}
//-----------------------------------------------------------------------------
