// Copyright (C) 2010 Johan Hake
// Licensed under the GNU LGPL Version 3.0
//
// First added:  2010-01-19
// Last changed: 2012-05-02

#ifndef __MARKOVMODEL_H
#define __MARKOVMODEL_H

#include <string>
#include <boost/function.hpp>
#include <boost/signal.hpp>

#include "common.h"

namespace gillstep
{

  class GillespieStepper;
  
  /// The base class of generated Markov Models
  class MarkovModel
  {
  public:
    
    /// Default constructor
    MarkovModel(const std::string& name, GillespieStepper& gillstepper);
    
    /// Destructor
    virtual ~MarkovModel();

    /// Initialize (reset) the Markov model using the present external values
    void reset();

    /// Return the name of the Markov model
    const std::string name() const {return _name;}
    
    /// Returns the maximal transitions from a state
    virtual uint max_transitions() const = 0;

    /// Return the 'value' of the Markov model or its parent model if this is a SubModel
    virtual double value() const = 0;

    /// Return the size of the substep. 1.0 if no substep is needed
    virtual double substep_size(double present_substep) = 0;
    
    /// Evaluate rate function corresponding to passed transition
    virtual double evaluate_rate(uint transition) const = 0;

    /// Evaluate rate function corresponding to passed state and transition
    virtual double evaluate_rate(uint state, uint transition) const = 0;

    /// Evaluate rate function corresponding to passed transition and a give 
    /// substep
    virtual double evaluate_substep_rate(uint transition, double present_substep, 
					 double next_substep) = 0;

    /// Change the state to the given transition, return true if value is changed
    virtual bool change_state(int transition) = 0;
    
    /// Get parameter
    virtual double get_parameter(const std::string& name) const = 0;
    
    /// Set parameter
    virtual void set_parameter(const std::string& name, double value) = 0;

    /// Get parameter names
    virtual std::vector<std::string> get_parameter_names() const  = 0;

    /// Return the number of states
    virtual uint num_states() const = 0;

    /// Return the number of sub models
    inline uint num_submodels() const 
    { return _submodels.size(); }

    /// Update the temperature scale factor
    virtual void update_temp_scale() = 0;

    /// Is a parent connected to its submodels
    inline bool receive_parent() const 
    { return _receive_parent; }
    
    /// Return the i:th sub model
    inline MarkovModel& submodel(uint i) 
    {
      if (i < num_submodels())
	return *_submodels[i];
      throw std::runtime_error("No submodel with that number excists");
    }

    /// Return the i:th sub model (const version)
    inline const MarkovModel& submodel(uint i) const 
    {
      if (i < num_submodels())
	return *_submodels[i];
      throw std::runtime_error("No submodel with that number excists");
    }

    /// Return true if Markov model is a sub model
    inline bool is_submodel() const 
    { return _parent != 0; }

    /// Return reference to parent, if not parent return reference to self
    inline MarkovModel& get_parent() 
    { return is_submodel() ? *_parent : *this ; }
    
    /// Return the state
    uint state() const 
    { return _state; }

    /// Connect slot to the signal
    void connect(ValueReceiver slot)
    { _signal.connect(slot); }
    
    /// Send the value of the Markov model to any registered slots
    void send() const 
    //    { _signal(value()); std::cout << "sending: " << name() << ", " << value() << std::endl;}
    { _signal(value()); }

    /// Update the continuous variables to a certain substep
    void update_variables(double substep);

    /// Return the value of a variable
    double get_variable(std::string name) const;

    /// Set the value of a variable
    void set_variable(std::string name, double value);

    /// Return a ValueReceiver used to receive a particular variable
    ValueReceiver receive(std::string variable);

    /// Return true if a variable can be received by the Markov Model
    bool can_receive(std::string name) const;

    /// Return a vector of variables names
    std::vector<std::string> variable_names() const;

    /// Return the value of the ith variable (fast unsafe and hidden function)
    inline double _unsafe_variable(uint i) const
    { return _present_variables->operator[](i); }

    /// Return the value of the ith discrete variable 
    /// (fast unsafe and "hidden" function)
    inline double _unsafe_discrete_variable(uint i) const
    { return _discrete_variables[i]; }

    /// An attribute that can be set to controll initialization of a Markov Model
    int init_state;

    /// An attribute that controlls if the Markov model should be fixed to initial
    /// value
    bool fix_state;

  protected:

    /// Variable difference
    bool _retabulate_derivatives;

    /// The state value
    uint _state;

    /// Old value
    double _old_value;

    /// Vector with external variables
    std::vector<double>* _present_variables;
    std::vector<double> _variables_0;
    std::vector<double> _variables_1;
    std::vector<double> _discrete_variables;

    /// Map of variable name to indices
    std::map<std::string, uint> _variable_map;
    std::map<std::string, uint> _discrete_variable_map;

    /// Map of ValueReceiver slots
    std::map<std::string, ValueReceiver> _value_receivers;

    /// Pointers to sub Markov models
    std::vector<MarkovModel*> _submodels;

    /// The gillespie stepper which the Markov model is registered with
    GillespieStepper* _gillstepper;

    /// A pointer to a potential parent Markov model
    MarkovModel* _parent;
    
    /// Is a parent connected to its submodels
    bool _receive_parent;

  private:
    
    /// The name of the Markov model
    std::string _name;

    /// The boost::signal, used to communicate the value of the Markov model
    boost::signal<void (double)> _signal;

  };

  /// Rate function call
  typedef boost::function<double (const MarkovModel& mm)> rate_call;

}

#endif
