__author__ = "Johan Hake (hake.dev@gmail.com)"
__copyright__ = "Copyright (C) 2010 " + __author__
__date__ = "2010-01-26 -- 2012-04-23"
__license__  = "GNU LGPL Version 3.0 or later"

# Import statements
from gillstep.codesnippets import *
from gillstep.config       import config
from gillstep.modeleditor  import ModelEditor

__all__ = ["generate_and_write_code"]

def generate_and_write_code(mm, return_files_names=False):
    "Generates and write code to file"
    if not isinstance(mm, ModelEditor):
        raise TypeError, "expected a ModelEditor"
    codes = []
    files = []
    if mm.submodels():
        unique_sm = []
        for sm in mm.submodels():
            if not sm.markov_model() in unique_sm:
                codes.append(generate_model_code(sm.markov_model()))
                unique_sm.append(sm.markov_model())

    codes.append(generate_model_code(mm))
    code = "\n\n".join(codes)
    
    # Generate code and write to file
    filename = mm.name() + ".h"
    files.append(filename)
    out = open(filename, "w")
    out.write(code)
    out.close()

    if return_files_names:
        return code, files
    return code

def generate_model_code(model):
    "Generates the markov model code"
    assert(isinstance(model, ModelEditor))
    model_form["submodel_friends"]    = submodel_friends(model)
    model_form["class_name"]          = model.name()
    model_form["num_states"]          = model.rates.num_states()
    model_form["num_variables"]       = len(model.continuous_variables())
    model_form["num_discrete_variables"] = len(model.discrete_variables())
    model_form["init_params"]         = init_params_code(\
                                                model.parameter_defaults())
    model_form["state_values"]        = state_value_code(\
                                                model.state_values)
    model_form["init_rate_functions"] = init_rate_function_code(\
                                                model.rates)
    model_form["init_derivative_of_rate_functions"] \
                                             = init_derivative_of_rate_function_code\
                                               (model)
    model_form["init_transition_map"] = init_transition_map_code(\
                                                model.rates)
    model_form["init_param_map"]      = init_param_map_code(\
                                                model.parameter_defaults())
    model_form["num_parameters"]      = model.num_vars("parameters")
    model_form["parameter_names"]     = parameter_names(\
                                               model.parameter_defaults())
    model_form["max_transitions"]     = model.rates.max_number_of_transitions()
    model_form["define_rate_functions"] \
                                             = define_rate_functions_code(model)
    model_form["define_derivatives_of_rate_functions"] \
                                             = define_derivatives_of_rate_functions_code(model)
    model_form["submodel_data"] \
                                             = submodel_data_code(model)
    receive_methods, receive_objects_init    = receive_methods_code(model)
    model_form["receive_methods"]      = receive_methods
    model_form["receive_objects_init"] = receive_objects_init
    model_form["submodels_cstr_1"] \
                                             = submodels_cstr_code_1(model)
    model_form["submodels_cstr_2"] \
                                             = submodels_cstr_code_2(model)
    model_form["evaluate_rate"]       = evaluate_rate_code(model)

    file_form["model_code"] = model_snippet%model_form
    file_form["CLASS_NAME"]        = model.name().upper()
    # FXIME: Include the whole code directly!
    file_form["includes"]          = ""#includes_code(model)
    
    return file_snippet%file_form

def indent(s, n):
    "Indent each row of the given string s with n spaces"
    indentation = " "*n
    return indentation + ("\n" + indentation).join(s.split("\n"))

def indent_and_join(l, n):
    """Indent each row of the given list of strings l with n spaces 
    and return a joined string"""
    indentation = " "*n
    return indentation + ("\n" + indentation).join(l)

init_code_indent = 6

def init_params_code(parameter_defaults):
    return indent_and_join(\
        ["params_.push_back(%f);"%value\
         for value in parameter_defaults.itervalues()], init_code_indent)

def parameter_names(parameter_defaults):
    return indent_and_join(\
        ["parameter_names.push_back(\"%s\");"%value\
         for value in parameter_defaults.iterkeys()], init_code_indent)

def state_value_code(state_values):
    state_values_ = []
    for state in state_values:
        state_values_.append("  case %d:"%state)
        state_values_.append("    return %s;"%state_values.to_ccode(state).replace("mm.",""))
    
    return indent_and_join(state_values_, init_code_indent)

def define_rate_functions_code(model):
    rate_function_code = []
    function_template="""double rate_func_%d_%d(const MarkovModel& mm)
{
  const double value = %s::temp_scale()*(%s);
//  std::cout << mm.name() << " tau: %d-%d: " << 1/value << std::endl;
  return value;
}
"""
    for con in model.rates:
        rate_function_code.append("")
        rate_function_code.append("/// Rate function between state: %d and %d "%con)
        rate_function_code.extend((function_template%(con[0], con[1],\
                                model.name(), model.rates.to_ccode(con), \
                                con[0], con[1])).split("\n"))

    rate_function_code.append("")
    
    return indent_and_join(rate_function_code, init_code_indent-2)

def define_derivatives_of_rate_functions_code(model):
    rate_function_code = []
    function_template="""double der_rate_func_%d_%d_wrt_%s(const MarkovModel& mm)
{
  return %s::temp_scale()*(%s);
}
"""
    for con in model.rates:
        for var, code in model.rates.derivative_ccode(con):
            rate_function_code.append("")
            rate_function_code.append("/// Derivative of rate function between state: %d and %d for variable: %s"%(con[0], con[1], var))
            rate_function_code.extend((function_template%(con[0], con[1], var, 
                         model.name(), code)).split("\n"))

    rate_function_code.append("")
    return indent_and_join(rate_function_code, init_code_indent-2)
    
def init_derivative_of_rate_function_code(model):
    rate_function_code = []
    
    # Push the derivative of rate functions to the vectors
    for from_state in xrange(model.rates.num_states()):
        rate_function_code.append("")
        rate_function_code.append("// Derivative of rate functions from state %d"%\
                                  from_state)
        rate_function_code.append("std::vector<std::vector<rate_call> > der_rate_functions_%d;"%from_state)
        for to_state in model.rates.state_connections()[from_state]:
            
            rate_function_code.append("")
            rate_function_code.append("// Derivative of rate functions from state %d to state %d"%(from_state, to_state))
            rate_function_code.append("std::vector<rate_call> der_rate_functions_%d_%d;"%(from_state, to_state))
            for var in model.continuous_variables().keys():
                rate_function_code.append("der_rate_functions_%d_%d.push_back(der_rate_func_%d_%d_wrt_%s);"%(from_state, to_state, from_state, to_state, var))
            
            rate_function_code.append("der_rate_functions_%d.push_back(der_rate_functions_%d_%d);"%(from_state, from_state, to_state))

        rate_function_code.append("")
        rate_function_code.append(\
            "der_rate_functions_.push_back(der_rate_functions_%d);"%from_state)
        
    return indent_and_join(rate_function_code, init_code_indent)

def init_rate_function_code(rates):
    rate_function_code = []
    
    # Push the rate functions to the vectors
    for from_state in xrange(rates.num_states()):
        rate_function_code.append("")
        rate_function_code.append("// Rate functions from state %d"%from_state)
        rate_function_code.append(\
            "std::vector<rate_call> rate_functions_%d;"%from_state)
        for to_state in rates.state_connections()[from_state]:
            rate_function_code.append(\
                "rate_functions_%d.push_back(rate_func_%d_%d);"%(from_state, \
                from_state, to_state))

        rate_function_code.append(\
            "rate_functions_.push_back(rate_functions_%d);"%from_state)
        
    return indent_and_join(rate_function_code, init_code_indent)

def init_transition_map_code(rates):
    transition_map_code = []

    # Push the RateFunctions map to the vectors
    for from_state in xrange(rates.num_states()):
        transition_map_code.append("")
        transition_map_code.append("// Transition map from state %d"%from_state)
        transition_map_code.append(\
            "std::vector<uint> transition_map_%d;"%from_state)
        for to_state in rates.state_connections()[from_state]:
            transition_map_code.append(\
                "transition_map_%d.push_back(%d);"%(from_state, to_state))
        
        transition_map_code.append(\
            "transition_map_.push_back(transition_map_%d);"%from_state)
            
    return indent_and_join(transition_map_code, init_code_indent)

def includes_code(model):
    if len(model.submodels()) == 0:
        return ""

    included = []
    includes = [""]
    includes.append("/// Add declaration/definitions of Sub models")
    for sm in model.submodels():
        # Only include once
        if sm.markov_model() in included:
            continue
        includes.append("#include \"%s.h\""%sm.markov_model().name())
        included.append(sm.markov_model())
    includes.append("")

    return indent_and_join(includes, 0)

def receive_methods_code(model):
    receive_methods = []
    receive_objects_init = []
    
    if not model.continuous_variables() and not model.discrete_variables():
        return "", ""

    receive_methods.append("")
    receive_methods.append("/// Receive methods")
    for i, var in enumerate(model.continuous_variables().iterkeys()):
        receive_objects_init.append("_value_receivers[\"%s\"] = boost::bind(&%s::_receive_%s, this, _1);"%(var, model.name(), var))
        receive_objects_init.append("_variable_map[\"%s\"] = %d;"%(var, i))
        receive_methods.append("inline void _receive_%s(double value)"%var)
        receive_methods.append("{")
        #receive_methods.append("  std::cout << this->name() << \" received \" << \"%s%i: \" << value << std::endl;"%(var, i))
        receive_methods.append("  _retabulate_derivatives = true;")
        receive_methods.append("  _variables_0[%d] = _variables_1[%d];"%(i, i))
        receive_methods.append("  _variables_1[%d] = value;"%i)
        receive_methods.append("}")
        receive_methods.append("")

    for i, var in enumerate(model.discrete_variables().iterkeys()):
        receive_objects_init.append("_value_receivers[\"%s\"] = boost::bind(&%s::_receive_%s, this, _1);"%(var, model.name(), var))
        receive_objects_init.append("_discrete_variable_map[\"%s\"] = %d;"%(var, i))
        receive_methods.append("inline void _receive_%s(double value)"%var)
        receive_methods.append("{")
        #receive_methods.append("  std::cout << this->name() << \" received \" << \"%s%i: \" << value  << std::endl;"%(var, i))
        receive_methods.append("  _retabulate_derivatives = true;")
        receive_methods.append("  _discrete_variables[%d] = value;"%i)
        receive_methods.append("}")
        receive_methods.append("")

    return indent_and_join(receive_methods, init_code_indent-2), \
           indent_and_join(receive_objects_init, init_code_indent)

def init_param_map_code(parameter_defaults):
    return indent_and_join(\
        ["param_map_[\"%s\"] = %d;"%(name, i)\
         for i, name in enumerate(parameter_defaults.iterkeys())], init_code_indent)

def submodels_cstr_code_1(model):
    if not model.submodels():
        return ""
    submodel_cstr = []
    for sm in model.submodels():
        submodel_cstr.append(", %s(this, gillstepper)"%(str(sm)))

    return "".join(submodel_cstr)

def submodels_cstr_code_2(model):
    if not model.submodels():
        return ""
    submodel_cstr = []
    for sm in model.submodels():
        submodel_cstr.append("_submodels.push_back(&%s);"%(str(sm)))
        
    return indent_and_join(submodel_cstr, init_code_indent)

def submodel_friends(model):
    submodel_friends = []
    if not model.submodels():
        submodel_friends.append("")
        return indent_and_join(submodel_friends, init_code_indent)
    submodel_friends.append("")
    submodel_friends.append("/// Sub models friends")
    already_added = []
    for sm in model.submodels():
        class_name = sm.markov_model().name()
        if class_name in already_added:
            submodel_friends.append("friend class %s;"%(class_name))
            already_added.append(class_name)
    submodel_friends.append("")
    return indent_and_join(submodel_friends, init_code_indent-2)
    

def submodel_data_code(model):
    submodel_data = []
    if not model.submodels():
        submodel_data.append("")
        return indent_and_join(submodel_data, init_code_indent)

    submodel_data.append("")
    submodel_data.append("/// Sub models")
    for sm in model.submodels():
        submodel_data.append("%s %s;"%(sm.markov_model().name(), str(sm)))
    submodel_data.append("")
    return indent_and_join(submodel_data, init_code_indent-2)

def evaluate_rate_code(model):
    evaluate_rate = []
    
    for from_state in xrange(model.rates.num_states()):
        evaluate_rate.append("  case %d:"%from_state)
        evaluate_rate.append("    switch (to_state) {")
        trans = 0
        for to_state in xrange(model.rates.num_states()):
            if (from_state, to_state) in model.rates:
                evaluate_rate.append("      case %d:"%to_state)
                evaluate_rate.append("        return %s::_rate_functions"\
                                     "[from_state][%d](*this);"%(model.name(), trans))
                trans += 1

        evaluate_rate.append("  default:")
        evaluate_rate.append("    return 0.0;")
        evaluate_rate.append("  }")

    return indent_and_join(evaluate_rate, init_code_indent)
