"This module provides functionality to compile PyGILLSTEP compatible extension modules."

__author__ = "Johan Hake (hake.dev@gmail.com)"
__copyright__ = "Copyright (C) 2010 " + __author__
__date__ = "2009-08-15 -- 2010-09-02"
__license__  = "GNU LGPL Version 3.0 or later"

import sys
import os
import re
import numpy
import instant
import hashlib
import types

# Set log level of instant
instant.set_logging_level("WARNING")

# Import PyGILLSTEP
import gillstep

__all__ = ["jit"]

_additional_declarations = r"""
%%init%%{
import_array();
%%}

%%include <exception.i>
%%include <std_string.i>
%%include <gillstep/swig/typemaps.i>
%%include <gillstep/swig/exceptions.i>
%%ignore get_parameter(uint ind);

%(additional_declarations)s

%%pythoncode %%{
import gillstep.cpp
'''
A hack to get passed a bug in swig.
This is fixed in swig version 1.3.37
%%}
%(gillstep_import_statement)s
%%pythoncode %%{
'''
%%}
"""

def jit(markov_model, gillstepper, num=1):
    "JIT compile a Markov model"
    if not isinstance(markov_model, gillstep.ModelEditor):
        raise TypeError, "expected a 'ModelEditor' as first argument"

    if not isinstance(gillstepper, gillstep.cpp.GillespieStepper):
        raise TypeError, "expected a 'GillespieStepper' as second argument"

    # Generate the code
    code = gillstep.codegeneration.generate_and_write_code(markov_model)

    # Read code from file
    #code = open(markov_model.name() + ".h").read()

    # Read in code from any submodels
    submodel_code = ""
    #distinct_submodels = []
    #for sm in markov_model.submodels():
    #    if sm.markov_model().name() not in distinct_submodels:
    #        distinct_submodels.append(sm.markov_model().name())
    #        submodel_code += open(sm.markov_model().name() + ".h").read()

    # Compile module
    compiled_module = compile_extension_module(code, submodel_code)

    # Extract subclass
    SubModel = getattr(compiled_module, markov_model.name())
    assert(issubclass(SubModel, gillstep.cpp.MarkovModel))

    # Add a better repr method
    SubModel.__repr__ = types.MethodType(lambda self : gillstep.cpp.MarkovModel.__repr__(self),\
                                         None, SubModel)
    
    # Clean generated code
    #for f in files:
    #    os.unlink(f)

    # Instantiate the sub class
    if num == 1:
        return SubModel(gillstepper)
    return [SubModel(gillstepper) for i in xrange(num)]

def compile_extension_module(code, submodel_code="", module_name="",
                             recompile=False,
                             additional_declarations="",
                             additional_system_headers=None):
    """ Compile a PyGILLSTEP compatible extension module, based on C++ GILLSTEP code"""
    # Check the provided arguments
    expect_arg(str, code, "first")
    expect_arg(str, module_name, "module_name")
    expect_arg(str, additional_declarations, "additional_declarations")
    expect_arg(bool, recompile, "recompile")
    additional_system_headers = \
                expect_list_of(str, additional_system_headers, "additional_system_headers")

    # Check that the code does not use 'using namespace gillstep'
    if re.search("using[ \n]+namespace[ \n]+gillstep",code):
        raise ValueError, "Do not use 'using namespace gillstep'. Include the code in namespace gillstep {...} instead."

    # Check that the code is included in gillstep namespace
    if not re.search("namespace[ \n]+gillstep[ \n]{",code):
        raise ValueError, "Include the C++ code in the 'gillstep namspace', otherwise SWIG cannot figure out the right PyGILLSTEP types."

    # Create unique module name for this application run
    if module_name is "":
        module_name = "gillstep_compile_code_%s" % hashlib.md5(repr(code) +\
                                            repr(submodel_code) +\
                                            instant.get_swig_version() +\
                                            gillstep.__version__).hexdigest()
    # Check cache
    if not recompile:
        compiled_module = instant.import_module(module_name)

        if compiled_module:
            return compiled_module
    gillstep.logger.push_level(gillstep.logger.INFO)
    gillstep.logger.info("Calling GILLSTEP just-in-time (JIT) compiler, this may take some time...")
    sys.stdout.flush()
    # Configure instant and add additional system headers
    instant_kwargs = configure_instant()
    instant_kwargs["system_headers"] += additional_system_headers
    instant_kwargs['include_dirs'] += [os.path.abspath(os.path.curdir)]

    declaration_strs = {"additional_declarations":""}
    using_reduced_import = False
    declaration_strs["gillstep_import_statement"] = r"%import gillstep/swig/gillstep.i"

    # Add any provided additional declarations
    if additional_declarations is not None:
        declaration_strs["additional_declarations"] += additional_declarations

    # Compile extension module with instant
    compiled_module = instant.build_module(\
             code              = code,
             additional_declarations = _additional_declarations%declaration_strs,
             signature         = module_name,
             **instant_kwargs)

    gillstep.logger.info(" done")
    gillstep.logger.pop_level()
    sys.stdout.flush()
    return compiled_module

def get_gillstep_include_dir():
    # Get system configuration
    (gillstep_include_dirs,
     dummy,
     dummy,
     dummy) = instant.header_and_libs_from_pkgconfig("gillstep")

    # Let swig see the installed gillstep swig files
    gillstep_include_found = False
    for inc_dir in gillstep_include_dirs:
        # Check if gillstep.h is installed in gillstep_include_dir
        if os.path.exists(os.path.join(inc_dir, "gillstep.h")):
            return os.path.abspath(inc_dir)
    raise OSError, "Could not find 'gillstep.h', make sure it is installed in a directory that 'pkg-config --cflags gillstep' returns."

def configure_instant():
    """Check system requirements

    Returns a dict with kwargs that can be passed to instant.build_module.
    """
    instant_kwargs = {}
    # Get system configuration
    (instant_kwargs['include_dirs'],
     flags,
     instant_kwargs['libraries'],
     instant_kwargs['library_dirs']) = instant.header_and_libs_from_pkgconfig("gillstep")

    # Let swig see the installed gillstep swig files
    swig_include_dirs = []
    gillstep_include_found = False
    ufc_include_found = False
    for inc_dir in instant_kwargs['include_dirs']:
        # Check if gillstep is installed in inc_dir
        if os.path.exists(os.path.join(inc_dir, "gillstep", "swig", "gillstep.i")):
            gillstep_include_found = True
            if inc_dir not in swig_include_dirs:
                swig_include_dirs.append(inc_dir)

    if not gillstep_include_found:
        raise OSError, """Didn't find gillstep.i in include paths returned from pkg-config.
Please make sure that your GILLSTEP installation corresponds with the one returned by pkg-config"""

    # Check that the form compiler will use the same swig version
    # that PyGILLSTEP was compiled with
    if not instant.check_swig_version(gillstep.cpp.__swigversion__, same=True):
        raise OSError, """GILLSTEP was not compiled with the present version of swig.
Install swig version %s or recompiled GILLSTEP with present swig
"""%gillstep.cpp.__swigversion__

    instant_kwargs['swig_include_dirs'] = swig_include_dirs
    instant_kwargs['include_dirs'].append(numpy.get_include())
    instant_kwargs['system_headers'] = ["cmath", "iostream",
                                        "stdexcept","numpy/arrayobject.h",
                                        "gillstep.h","boost/shared_ptr.hpp"]
    instant_kwargs['swigargs'] =['-c++','-I.', '-O']

    instant_kwargs['cppargs'] = []

    return instant_kwargs

def expect_list_of(argtype,arg,argname):
    assert(isinstance(argtype,type))
    error_str = "Provide a 'tuple' or 'list' with '%s', for the '%s' argument"%(argtype.__name__,argname)
    if arg is None:
        return []
    if not isinstance(arg,(types.NoneType,list,tuple)):
        raise TypeError, error_str
    if not all(isinstance(s, str) for s in arg):
        raise TypeError, error_str
    return arg

def expect_arg(argtype,arg,argname):
    if isinstance(argtype,(tuple,list)):
        all(isinstance(s, type) for s in argtype)
    else:
        assert(isinstance(argtype,type))
        type_str = "'%s'"%argtype.__name__

    # Check the type of the argument
    if not isinstance(arg,argtype):
        if isinstance(argtype,(tuple,list)):
            if len(argtype) > 2:
                argtype_str = ", ".join("'%s'"%s.__name__ for s in argtype[:-1]) + " or " + "'%s'"%argtype[-1]
            else:
                argtype_str = " or ".join("'%s'"%s.__name__ for s in argtype)
        else:
            argtype_str = "'%s'"%s.__name__
        raise TypeError, "Expected a '%s' for the '%s' argument"%argname

