__author__ = "Johan Hake (hake.dev@gmail.com)"
__copyright__ = "Copyright (C) 2010 " + __author__
__date__ = "2010-03-10 -- 2011-02-22"
__license__  = "GNU LGPL Version 3.0 or later"

__all__ = ["load_model", "load_submodel"]

# Import gillstep modules
import gillstep
from   gillstep.namespace   import gillstep_namespace
from   gillstep.modeleditor import ModelEditor
from   gillstep.submodel    import SubModel, to_submodels
from   gillstep.runtime     import update_runtime_namespace
from   gillstep.logger      import error

# Empty dict that will be populated with optional arguments
_arguments = {}

def arguments(**kwargs):
    """Called from within a Markov model form file to set arguments that can
    be changed by an external caller"""
    global _arguments

    # Check if any external arguments for this model has been set
    for var in _arguments:
        if var not in kwargs:
            raise RuntimeError, "Can only set variables that are defined using 'arguments(foo='bar') in the form file.'"
    
    # Populate the namespace first with passed arguments
    __populate_caller_namespace(kwargs)

    # and then with any external arguments stored in _arguments
    __populate_caller_namespace(_arguments)
        
def load_model(name, **kwargs):
    """Load a Markov model from file and return the instance

The method looks for a file with .mm extension.

External variables that can be read into the form file can be
passed through kwargs
"""

    global _arguments

    # Update the _argument dict
    _arguments.update(kwargs)

    # Create the markov model editor which shall be populated by the form file
    mm = gillstep.ModelEditor(name)

    # Get namespace from Markov model and populate it with gillstep namespace
    ns = mm.namespace()
    ns.update(gillstep_namespace())
    update_runtime_namespace(mm, ns)
    ns['load_submodel'] = load_submodel
    ns['arguments'] = arguments

    # Read and execute Markov model form file
    execfile(name + ".mm", ns)

    # Analyse the Markov model
    if not mm.analyze():
        error("Error while executing '%s'..."%name, raise_error=True)

    # Remove any external variables
    _arguments.clear()

    return mm

def load_submodel(name, number=1, **kwargs):
    "Load a Markov model to be used in Model declaration"
    # Load model from file
    mm = load_model(name, **kwargs)

    # Return a list of SubMarkovModels
    return to_submodels(mm, number)


def __populate_caller_namespace(ns):
    "Populate callers namespace"
    assert(isinstance(ns, dict))
    
    import inspect
    

    # We need to go two frames back, as this method should be called from
    # within load markov models above.
    frame = inspect.currentframe().f_back.f_back
    try:
        for key, value in ns.iteritems():
            frame.f_globals[key] = value
    finally:
        del frame
        
