__author__ = "Johan Hake (hake.dev@gmail.com)"
__copyright__ = "Copyright (C) 2010 " + __author__
__date__ = "2010-01-26 -- 2011-08-15"
__license__  = "GNU LGPL Version 3.0 or later"

__all__ = ["ModelEditor"]

# Import statements
import sympy

# Import gillstep modules
from gillstep.ordereddict import OrderedDict
from gillstep.logger      import error, info, debug
from gillstep.namespace   import T
from gillstep.gillstepsymbol import GillstepSymbol

class Rates(dict):
    "A help class to keep track of the rates"
    def __init__(self, mm):
        if not isinstance(mm, ModelEditor):
            error("expected a ModelEditor as 'mm' argument", raise_error=True)

        # Init base class
        dict.__init__(self)

        # Keep a count for each added rate function
        self._count = 0

        # Store parent model
        self._mm = mm

        self._state_connections = None

    def __setitem__(self, states, expr):
        # A small check for the kind of keys we accept
        self._count += 1

        # Check for valid connections
        if not (isinstance(states, tuple) and len(states) == 2
                and all(isinstance(state, int) for state in states)):
            error("expected two integers defining direction of rate as key "\
                  "for rate %d: %s"%(self._count, str(states)))
            return 

        # Check for valid rate expression
        if not isinstance(expr, (sympy.Basic, int, float)):
            error("expected scalar or SymPy expression for rate %d: '%s'"%\
                  (self._count, str(func)))
            return

        # If scalar
        if isinstance(expr, (int, float)):
            expr = GillstepSymbol(str(expr))

        # Check for SubModelValues and SubModelParameters
        for smv in get_submodel_values(expr):
            self._mm._add_submodel_value(smv)

        for smp in get_submodel_parameters(expr):
            self._mm._add_submodel_parameter(smp)

        # Set the item
        dict.__setitem__(self, states, expr)

    def analyze(self):
        "Analyze the registered states and raise an error if encountered"
        for con in self:
            if (con[1], con[0]) not in self:
                error("while analyzing 'rates': expected a rate from state "\
                      "%d to state %d"%(con[1],con[0]))
                return False
        return True

    def state_connections(self):
        "Return the state connections"
        if self._state_connections is None:
            self._state_connections = [[] for i in xrange(self.num_states())]
            connections = self.keys()
            connections.sort()
            for from_state, to_state in connections:
                self._state_connections[from_state] += [to_state]

        return self._state_connections

    def max_number_of_transitions(self):
        "Return the maximal number of transitions from any state"
        return max(len(con) for con in self.state_connections())

    def num_states(self):
        "Return the number of states"
        if self:
            return max(max(self)) + 1
        return 0

    def to_ccode(self, con):
        return sympy.ccode(self[con].subs(self._mm.code_replace()))

    def derivative_ccode(self, con):
        for var in self._mm.continuous_variables().itervalues():
            yield var, sympy.ccode(self[con].diff(var).subs(self._mm.code_replace()))

class StateValues(dict):
    "A help class to keep track of the state values"
    def __init__(self, mm):
        if not isinstance(mm, ModelEditor):
            error("expected a ModelEditor as 'mm' argument", raise_error=True)
        self._mm = mm
        dict.__init__(self)
    
    def __setitem__(self, state, value):
        # A small check for the kind of keys and values we accept
        if not (isinstance(state, int) and isinstance(value, \
                                    (int, float, sympy.Basic))):
            error("expected an integer as key and a scalar or SymPy "\
                  "expression as 'value'")
            return

        if isinstance(value, sympy.Basic):
            if not all(isinstance(a, (int, float, SubModelValue)) \
                       for a in value.atoms()):
                error("expected a SymPy expression consisting of only "\
                      "scalars, SubModelValues")
                return
            
            # Check for SubModelValues and SubModelParameters
            for smv in get_submodel_values(value):
                self._mm._add_submodel_value(smv)

        dict.__setitem__(self, state, value)

    def __getitem__(self, state):
        "Getitem defaults to 0"
        return self.get(state, 0.0)

    def analyze(self):
        "Check that the handed states values are sane"
        if len(self) == 0:
            error("while analyzing 'state_value': expected at least one state to have a value.")
            return False
        num_states = self.num_states()
        for state in self:
            if not (0 <= state and state < num_states):
                error("while analyzing 'state_value': state number must be in [0,%d]"%(num_states-1))
                return False
        return True

    def num_states(self):
        return self._mm.rates.num_states()

    def to_ccode(self, state):
        if isinstance(self[state], sympy.Basic):
            return sympy.ccode(self[state].subs(self._mm.code_replace()))
        return str(self[state])
    

class ModelEditor(object):
    " A class used to define a Markov model"
    def __init__(self, name):
        "Initialize the ModelEditor"
        if not isinstance(name, str):
            error("expected a 'str' as 'name' argument", \
                  raise_error=True)

        self._name = name
        self._submodels = []
        self.reset()

    def name(self):
        return self._name

    def submodels(self):
        "Return the submodels"
        return self._submodels

    def reset(self):
        "Reset the variables"
        
        # Dictionaries for keeping track of declared variables
        self._code_replace = {}
        self._ns = dict()
        self._parameter_defaults = OrderedDict()
        self._parameters = OrderedDict()
        self._externals = OrderedDict()
        self._external_fields = OrderedDict()
        self._cont_vars = OrderedDict()
        self._disc_vars = OrderedDict()

        # Expose the global parameter T and define T_ref and Q10
        self._ns['Q10']   = GillstepSymbol("Q10")
        self._ns['T_ref'] = GillstepSymbol("T_ref")
        self._ns['time']  = GillstepSymbol("time")
        
        # Public attributes the user needs to populate
        self.rates        = Rates(self)
        self.state_values = StateValues(self)

        self._parameter_defaults[self._ns['Q10']]   = 1.0
        self._parameter_defaults[self._ns['T_ref']] = 37.0

        self._parameters['Q10'] = self._ns['Q10']
        self._parameters['T_ref'] = self._ns['T_ref']

        # Code replace for global variables
        self._code_replace[T] = GillstepSymbol("get_T()")
        self._code_replace[self._ns['time']] = GillstepSymbol("_gillstepper->time();")
        
    def num_vars(self, name):
        " Return the number of variable of type 'name'"
        if not isinstance(name, str):
            error("expected a 'str' as 'name'")
            return

        allowed_names = ["parameters", "externals", "external_fields"]
        if name not in allowed_names:
            error("expected 'name' to be in %s"%(str(allowed_names)))
            return
            
        return len(eval("self._%s"%name))

    def set_parameters(self, **kwargs):
        "Register parameters in the Markov model"
        names = kwargs.keys()
        names.sort()
        offset = len(self._parameters)
        return_values = OrderedDict()

        for i, name in enumerate(names):
            value = kwargs[name]
            if not isinstance(value, (int, float)):
                error("expected scalars for parameter values")
                return
            
            par = GillstepSymbol(name)
            self._parameters[name] = par
            self._code_replace[par] = GillstepSymbol("%s::get_parameter(%d)"%\
                                                   (self._name,i+offset))
            #self._code_replace[par] = GillstepSymbol("%f"%value)
            self._parameter_defaults[par] = value
            self._ns[name] = par
            return_values[name] = par

        # Populate the namespace of the caller
        self.__populate_caller_namespace(return_values)
        
        # Return parameters
        if len(return_values) == 1:
            return return_values.values()[0]
        
        return tuple(return_values.values())

    def set_variables(self, *args, **kwargs):
        """Register external variables

Variables need to be passed as a list of strings containing the name of
the variables.

Optional kwargs are 'global' and 'discrete'. They can be True or False and
they defaults to False. If global is true the variable will be introduced
at the level of each GillespieStepper. (FIXME: Implement!*) If discrete is True
the the variable will be discrete and cannot be differentiable.
* Implement this in the GillespieStepper?
"""
        for kwarg in kwargs:
            if not kwarg in ["discrete", "global"]:
                raise RuntimeError("Got an unexpected kwarg '%s'"%kwarg)
        global_var = kwargs.get("global", False)
        if global_var:
            raise NotImplementedError
        discrete = kwargs.get("discrete", False)

        if len(set(args)) != len(args):
            error("no duplicates are allowed for variables")

        # Are we updating the discrete or the continuous variables
        if discrete:
            variables = self._disc_vars
            code_replace = "mm._unsafe_discrete_variable(%d)"
        else:
            variables = self._cont_vars
            code_replace = "mm._unsafe_variable(%d)"

        # Check for duplication of variable names
        if any(var in self._disc_vars or var in self._cont_vars
               for var in variables):
            raise ValueError, "registered a duplicated variable name"

        offset = len(variables)
        return_values = OrderedDict()

        for i, name in enumerate(args):
            if not isinstance(name, str):
                raise TypeError, "expected 'str' for variables"
        
            var = GillstepSymbol(name)
            variables[name] = var
            self._code_replace[var] = GillstepSymbol(code_replace%(i+offset))
            self._ns[name] = var
            return_values[name] = var

        # Populate the namespace of the caller
        self.__populate_caller_namespace(return_values)
        
        if len(return_values) == 1:
            return return_values.values()[0]
        
        return tuple(return_values.values())
    
    def set_Q10(self, value, T_ref):
        """ Set the Q10 value together with the reference temperatur
of the rates in the Markov model"""
    
        if not (isinstance(value, (int, float)) and isinstance(T_ref, (int, float))):
            error("expected two scalars for the Q10 values")
            return

        if not (value > 0 and 10 <= T_ref and value <= 40):
            error("expected a positive value for the Q10 value and a temperature between 10 and 40 degC")
    
        self._parameter_defaults[self._parameters['Q10']] = value
        self._parameter_defaults[self._parameters['T_ref']] = T_ref

    def parameters(self):
        "Return the declared parameters"
        return self._parameters
    
    def discrete_variables(self):
        "Return the declared discrete variables"
        return self._disc_vars

    def continuous_variables(self):
        "Return the declared discrete variables"
        return self._cont_vars

    def externals(self):
        "Return the declared externals"
        return self._externals

    def external_fields(self):
        "Return the declared external_fields"
        return self._external_fields

    def namespace(self):
        "Return namespace which code can be executed in"
        return self._ns
        
    def analyze(self):
        "Analyze the declared Model"

        if len(self.rates) < 2:
            error("expected at least two registered rates")

        if not (self.rates.analyze() and self.state_values.analyze()):
            return False

        # FIXME: It is actually usefull to not check for this!
        # FIXME: Make it parameterizable
        ## Check that all declared names is used in the rates
        #for name, p in self._parameters.iteritems():
        #    if name in ["Q10", "T_ref", "T"]:
        #        continue
        #    if all( p not in rate for rate in self.rates.itervalues()):
        #        error("the parameter '%s' was not used"%name)
        #        return False
        #
        #for name, ext in self._disc_vars.iteritems():
        #    if all( ext not in rate for rate in self.rates.itervalues()):
        #        error("the external '%s' was not used"%name)
        #        return False
        #
        #for name, ext in self._cont_vars.iteritems():
        #    if all( ext not in rate for rate in self.rates.itervalues()):
        #        error("the external '%s' was not used"%name)
        #        return False
        #
        #for name, ext in self._externals.iteritems():
        #    if all( ext not in rate for rate in self.rates.itervalues()):
        #        error("the external '%s' was not used"%name)
        #        return False
        #
        #for name, ext_field in self._external_fields.iteritems():
        #    if all(ext_field not in rate for rate in self.rates.itervalues()):
        #        error("the external '%s' was not used"%name)
        #        return False
        
        return True

    def parameter_defaults(self):
        "Return the collected parameter_defaults"
        return self._parameter_defaults

    def code_replace(self):
        "Return the collected code replacer"
        return self._code_replace

    def str(self, verbose):
        "Return a string representation of the Markov Model"
        if not verbose:
            return "Markov Model: '%s' with %d states and %d submodels"%\
                   (self.name(), self.rates.num_states(), len(self._submodels))
        return "Not Implemented"

    def __str__(self):
        "Return a pretty print representation of the Markov Model"
        return self.str(False)

    def _add_submodel_value(self, submodelvalue):
        "Add a submodel to this Markov model"
        if not isinstance(submodelvalue, SubModelValue):
            error("expected a SubModelValue as 'submodelvalue' argument'", \
                  raise_error=True)

        # Get the submodel
        submodel = submodelvalue.sm

        if submodel.markov_model().submodels():
            error("sub markov models cannot have submarkov models (yet...)", \
                  raise_error=True)
            
        if not submodel in self._submodels:
            self._submodels.append(submodel)
            ind = len(self._submodels)-1
        else:
            ind = self._submodels.index(submodel)

        # Add submodelvalue in code replace
        self._code_replace[submodelvalue] = GillstepSymbol(\
            "mm.submodel(%d).value()"%(ind))

    def _add_submodel_parameter(self, submodelparam):
        "Add a submodel to this Markov model"
        if not isinstance(submodelparam, SubModelParameter):
            error("expected a SubModelParameter as 'submodelparam' argument'", \
                  raise_error=True)

        # Get the submodel
        submodel = submodelparam.sm

        if submodel.markov_model().submodels():
            error("sub markov models cannot have submarkov models (yet...)", \
                  raise_error=True)
        
        if not submodel in self._submodels:
            self._submodels.append(submodel)
        
        # Add submodelparam in code replace
        self._code_replace[submodelparam] = GillstepSymbol(\
            "%s::get_parameter(%d)"%(submodel.markov_model().name(), \
                                     submodelparam.ind))

    def __populate_caller_namespace(self, ns):
        "Populate callers namespace"
        assert(isinstance(ns, dict))
        
        import inspect
        
        # We need to go two frames back, as this method should be called within
        # a method of the ModelEditor
        frame = inspect.currentframe().f_back.f_back
        try:
            for key, value in ns.iteritems():
                frame.f_globals[key] = value
        finally:
            del frame
        
# Import gillstep modules here to get around circular dependencies
from  gillstep.submodel import *
