__author__ = "Johan Hake (hake.dev@gmail.com)"
__copyright__ = "Copyright (C) 2010 " + __author__
__date__ = "2010-03-10 -- 2011-01-27"
__license__  = "GNU LGPL Version 3.0 or later"

import sympy

# Import gillstep modules
from gillstep.modeleditor import ModelEditor
from gillstep.logger      import error
from gillstep.gillstepsymbol import GillstepSymbol

__all__ = ["SubModel", "to_submodels", "get_submodel_values",
           "get_submodel_parameters", "SubModelParameter", "SubModelValue"]

class SubModelParameter(GillstepSymbol):
    "A wrapper for using a submodels parameters"
    def __new__(cls, sm, name):
        assert(isinstance(sm, SubModel))
        assert(isinstance(name, str))

        self = GillstepSymbol.__new__(cls, "%s.%s"%(str(sm), name))
        self.sm = sm
        self.ind = sm.markov_model().parameters().index(name)
        return self


class SubModelValue(GillstepSymbol):
    "A wrapper for using a submodels values"
    def __new__(cls, sm):
        assert(isinstance(sm, SubModel))
        self = GillstepSymbol.__new__(cls, "%s.value"%(str(sm)))
        self.sm = sm
        return self

    def __call__(self, *arg, **kwargs):
        raise TypeError,"cannot call a 'SubModelvalue'"

class SubModel(object):
    "A wrapper around a Markov model when such is used as a SubModel"
    _mm_dict = {}
    def __init__(self, mm):
        assert(isinstance(mm, ModelEditor))
        # Keep track of all SubModels
        if mm not in SubModel._mm_dict:
            SubModel._mm_dict[mm] = {}
            SubModel._mm_dict[mm][self] = 1
        else:
            SubModel._mm_dict[mm][self] = len(SubModel._mm_dict[mm])+1
        
        # Store the ModelEditor
        self._mm = mm

        # Store attributes that are available for declaring MarkovModels
        self.value = SubModelValue(self)
        for param in mm.parameters():
            setattr(self, param, SubModelParameter(self, param))
        
    def markov_model(self):
        "Return the wrapped Markov model"
        return self._mm

    def __str__(self):
        return "sub_%s_%d"%(self._mm.name(), SubModel._mm_dict[self._mm][self])

def to_submodels(mm, number=1):
    """Convert a ModelEditor into a number of SubModels
to be used in Model declarations"""
    if not isinstance(mm, ModelEditor):
        error("expected a ModelEditor for 'mm' argument", raise_error=True)

    if not isinstance(number,int) or number < 1:
        error("expected a positive integer for 'number' arguement", raise_error=True)

    return SubModel(mm) if number ==1 else tuple(SubModel(mm) for i in xrange(number))

def get_submodel_values(expr):
    """Takes a Sympy expression and check if there are any SubModels in it
if so return a list of them.
"""
    if not isinstance(expr, sympy.Basic):
        error("expected a sympy expression for 'expr' argument")
    return sorted([s for s in expr.atoms() if isinstance(s, SubModelValue)])

def get_submodel_parameters(expr):
    """Takes a Sympy expression and check if there are any SubModels in it
if so return a list of them.
"""
    if not isinstance(expr, sympy.Basic):
        error("expected a sympy expression for 'expr' argument")
    return sorted([s for s in expr.atoms() if isinstance(s, SubModelParameter)])
