__author__ = "Johan Hake (hake.dev@gmail.com)"
__date__ = "2010-05-27 -- 2010-09-02"
__copyright__ = "Copyright (C) 2010 " + __author__
__license__  = "GNU LGPL Version 3"

# Module to specify package tests This file should grow as more packages are added.

from scons_configure import get_package_dir, conf_message, get_status_output
from scons_tools import is_windows, is_darwin
import os, re, sys
def conf_unsuccessfull():
    print "Configuration was not successfull. \nLook into 'config.log' for details"
    sys.exit(1)

def check_swig(context):
    context.Message("Checking for SWIG...")
    command_str = "swig -version"
    conf_message(context.sconf, "Running: '%s'\n"%command_str)
    result, output = get_status_output(command_str)
    conf_message(context.sconf, "Result:\n%s\n"%output)
    context.Result(not result)
    if result != 0:
        conf_message(context.sconf, """SWIG not found.
Install SWIG or disable building of Python extension.
""")
        conf_unsuccessfull()
        
def check_swig_version(context, correct_version):
    context.Message("Checking for SWIG version >= %d.%d.%d..."%correct_version)
    command_str = "swig -version"
    conf_message(context.sconf, "Running: '%s'\n"%command_str)
    result, output = get_status_output(command_str)
    conf_message(context.sconf, "Result:\n%s\n"%output)
    swig_version = re.findall(r"SWIG Version ([0-9.]+)",output)[0]
    swig_enough = True
    for i, v in enumerate([int(v) for v in swig_version.split(".")]):
        if    correct_version[i] <= v  : break
        elif  correct_version[i] == v : continue
        else: swig_enough = False
    context.Result(swig_enough)
    if not swig_enough:
        conf_message(context.sconf, """SWIG version must be >= %d.%d.%d.
Install SWIG version >= %d.%d.%d or disable building of Python extension.
"""%(correct_version+correct_version))
        conf_unsuccessfull()

def check_python(context):
    context.Message("Checking for Python...")
    # Check for python development files
    python_version = "%s.%s" % (sys.version_info[0],sys.version_info[1])
    context.env["PYTHON_VERSION"] = python_version
    if is_windows(context.env):
        python_inc_dir = os.path.join(sys.prefix, "include")
        python_lib_dir = os.path.join(sys.prefix, "libs")
        python_lib = "python%s%s" % (sys.version_info[0], sys.version_info[1])
    else:
        python_inc_dir = os.path.join(sys.prefix, "include", "python"+python_version)

    python_is_found = os.path.isfile(os.path.join(python_inc_dir, "Python.h"))
    
    # Add path to python if found
    if python_is_found:
        context.env.Append(CPPPATH=[python_inc_dir])
        if is_windows(context.env):
            context.env.Append(LIBPATH=[python_lib_dir])
            context.env.Append(LIBS=[python_lib])
    context.Result(python_is_found)
    if not python_is_found:
        conf_message(context.sconf, """Unable to find Python development files on your system.
Perhaps you need to install the package python%s-dev?
Install Python development files or disable building of Python extension.
""" % swig_env["PYTHON_VERSION"])
        conf_unsuccessfull()


def check_boost(context):
    context.Message("Checking for Boost...")

    boost_found, result = get_package_dir("boost", context.env,
                                          os.path.join("boost","version.hpp"))

    # Add path if boost is found
    if boost_found:
        context.env.Append(CPPPATH = result)
        context.env['PACKAGE_DEPS']['CPPPATH'].append(result)
    
    context.Result(boost_found)
    if not boost_found:
        conf_message(context.sconf, result)
        conf_unsuccessfull()

def check_boost_signals(context):
    context.Message("Checking for Boost-signal...")

    found, result = get_package_dir("boost", context.env,
                                          header=os.path.join("boost","version.hpp"))

    if not found:
        conf_message(context.sconf, result, True, True)

    # If found add to path
    if found:
        context.env.Append(CPPPATH = result)
        context.env['PACKAGE_DEPS']['CPPPATH'].append(result)

    found, result = get_package_dir("boost", context.env,
                                          library="boost_signals")

    if found:
        # If found add to library path
        context.env.Append(LIBPATH = result, LIBS=["boost_signals"])
        context.env['PACKAGE_DEPS']['LIBPATH'].append(result)
        context.env['PACKAGE_DEPS']['LIBS'].append("boost_signals")
    else:
        found, result = get_package_dir("boost", context.env,
                                              library="boost_signals-mt")
        if found:
            # If found add to library path
            context.env.Append(LIBPATH = result, LIBS=["boost_signals-mt"])
            context.env['PACKAGE_DEPS']['LIBPATH'].append(result)
            context.env['PACKAGE_DEPS']['LIBS'].append("boost_signals-mt")
        else:
            conf_message(context.sconf, result, True, True)
    
    cpp_libs_str = r"""
#include <iostream>
#include <boost/signal.hpp>
#include <string>

void shout(std::string what){
  std::cout << what << std::endl;
}

int main() {
  boost::signal<void (std::string)> signal;
  signal.connect(shout);
  signal("Hello");
  return 0;
}
"""
    run_result, output = context.TryRun(cpp_libs_str, ".cpp")
    OK = run_result==1 and "Hello" in output
    context.Result(OK)

    if not OK:
        conf_unsuccessfull()


def check_boost_iostreams(context):
    context.Message("Checking for Boost-iostreams...")

    found, result = get_package_dir("boost", context.env,
                                          header=os.path.join("boost",
                                                              "iostreams",
                                                              "stream.hpp"))

    if not found:
        conf_message(context.sconf, result, True, True)

    # If found add to path
    if found:
        context.env.Append(CPPPATH = [result])
        context.env['PACKAGE_DEPS']['CPPPATH'].append(result)

    found, result = get_package_dir("boost", context.env,
                                          library="boost_iostreams")

    if found:
        # If found add to library path
        context.env.Append(LIBPATH = [result], LIBS=["boost_iostreams"])
        context.env['PACKAGE_DEPS']['LIBPATH'].append(result)
        context.env['PACKAGE_DEPS']['LIBS'].append("boost_iostreams")
    else:
        found, result = get_package_dir("boost", context.env,
                                              library="boost_iostreams-mt")
        if found:
            # If found add to library path
            context.env.Append(LIBPATH = [result], LIBS=["boost_iostreams-mt"])
            context.env['PACKAGE_DEPS']['LIBPATH'].append(result)
            context.env['PACKAGE_DEPS']['LIBS'].append("boost_iostreams-mt")
        else:
            conf_message(context.sconf, result, True, True)
    
    cpp_libs_str = r"""
#include <iostream>
#include <boost/iostreams/stream.hpp>
#include <boost/iostreams/filtering_stream.hpp>
#include <string>

int main() {
  boost::iostreams::filtering_stream<boost::iostreams::output> filter;
  std::cout << filter << "Hello";
  return 0;
}
"""
    run_result, output = context.TryRun(cpp_libs_str, ".cpp")
    OK = run_result==1 and "Hello" in output
    context.Result(OK)

    if not OK:
        conf_unsuccessfull()

def check_cppunit(context):
    context.Message("Checking for cppunit...")

    found, result = get_package_dir("cppunit", context.env,
                                    header=os.path.join("cppunit","TestResult.h"))

    if not found:
        conf_message(context.sconf, result, True, True)

    # If found add to path
    if found:
        context.env.Append(CPPPATH = [result])
        context.env['PACKAGE_DEPS']['CPPPATH'].append(result)

    found, result = get_package_dir("cppunit", context.env,
                                    library="cppunit")

    if found:
        # If found add to library path
        context.env.Append(LIBPATH = [result], LIBS=["cppunit"])
        context.env['PACKAGE_DEPS']['LIBPATH'].append(result)
        context.env['PACKAGE_DEPS']['LIBS'].append("cppunit")
    else:
        conf_message(context.sconf, result)
    
    cpp_libs_str = r"""
#include <cppunit/extensions/HelperMacros.h>
#include <cppunit/TestRunner.h>
#include <cppunit/TestResult.h>
#include <cppunit/TestResultCollector.h>
#include <cppunit/CompilerOutputter.h>

class Test : public CppUnit::TestFixture
{
  CPPUNIT_TEST_SUITE(Test);
  CPPUNIT_TEST(test_test);
  CPPUNIT_TEST_SUITE_END();
public:
  void test_test(){}
};

int main() {
  CppUnit::TestResult result;
  CppUnit::TestResultCollector collected_results;
  result.addListener(&collected_results);
  CppUnit::TestRunner runner;
  runner.addTest(CppUnit::TestFactoryRegistry::getRegistry().makeTest());
  runner.run(result);
  CppUnit::CompilerOutputter outputter(&collected_results, std::cout);
  outputter.write ();
  return collected_results.wasSuccessful () ? 0 : 1;
}
"""
    run_result, output = context.TryRun(cpp_libs_str, ".cpp")
    OK = run_result==1 and "OK" in output
    context.Result(OK)

    if not OK:
        conf_unsuccessfull()
