__author__ = "Johan Hake (hake.dev@gmail.com)"
__date__ = "2010-05-27 -- 2010-09-02" 
__copyright__ = "Copyright (C) 2010 " + __author__
__license__  = "GNU LGPL Version 3"

# Check for swig, swig version and boost installation
__all__ = ['configure_swig_env', 'configure_env']

import os, SCons, copy
from SCons.Script import *
from distutils  import sysconfig

def configure_swig_env(env, version, dependencies=None):
    " Configure and check swig dependencies"
    assert(isinstance(env, SCons.Script.Environment))

    # Return if help or cleaning
    if (env.GetOption('help') or env.GetOption('clean')):
        return prepare_swig_env(env)

    # Automatically handle dependencies
    dependencies = dependencies or []

    tests = {"check_swig"  : check_swig,
             "check_swigVersion" : check_swig_version,
             "check_python": check_python}

    for dep in dependencies:
        test = "check_" + dep
        try:
            tests[test] = eval(test)
        except:
            print "Configuration for '%s' is not found. Consider add it to the packages.py file under site_scons."%dep
            Exit(1)

    # Create configure environment
    conf = env.Configure(custom_tests=tests)

    # Call dependency checks
    conf_result = []
    
    conf.check_swig()
    conf.check_swigVersion(version)
    conf.check_python()
    
    for dep in dependencies:
        exec("conf.check_%s()"%dep)

    env = conf.Finish()

    # Return a prepared swig environment
    return prepare_swig_env(env)

def configure_env(env, dependencies = None, collect_dependencies=True):
    """ Configure and check swig dependencies """
    assert(isinstance(env, SCons.Script.Environment))

    # Check arguments
    if isinstance(dependencies, str):
        dependencies = [dependencies]
    elif not isinstance(dependencies, list):
        raise TypeError, "expected a list or a str for the dependencies argument"
    
    # Return if help or cleaning
    if (env.GetOption('help') or env.GetOption('clean')):
        return env

    # Cache dependencies
    package_deps = copy.deepcopy(env['PACKAGE_DEPS'])

    # Extend package dependencies
    env['PACKAGE_DEPS']['DEPENDENCIES'].extend(dependencies)
    
    # Automatically handle dependencies
    dependencies = dependencies or []

    tests = {}
    for dep in dependencies:
        test = "check_" + dep
        try:
            tests[test] = eval(test)
        except:
            print "Configuration for '%s' is not found. Consider add it to the packages.py file under site_scons."%dep
            Exit(1)

    # Create configure environment
    conf = env.Configure(custom_tests=tests)

    # Call dependency checks
    for dep in dependencies:
        exec("conf.check_%s()"%dep)

    env = conf.Finish()

    # If not collect dependencies erase any changes to package_deps
    if not collect_dependencies:
        env['PACKAGE_DEPS'] = package_deps

    return env

def prepare_swig_env(env):

    assert(isinstance(env, SCons.Script.Environment))
    
    # Change any optimization flags
    if "-O2" in env['CXXFLAGS']:
        env['CXXFLAGS'] = env['CXXFLAGS'].replace("-O2", "-O1")
    if "-O3" in env['CXXFLAGS']:
        env['CXXFLAGS'] = env['CXXFLAGS'].replace("-O3", "-O1")
    
    # Determine which compiler to be used:
    cxx_compilers = ["c++", "g++", "CC"]
    
    # Use CXX from os.environ if available:
    env["CXX"] = os.environ.get("CXX", env.Detect(cxx_compilers))
    
    if not env["CXX"]:
        print "Unable to find any valid C++ compiler."
        # try to use g++ as default:
        env["CXX"] = "g++"

    env["SHLIBPREFIX"] = "_"
    if env["PLATFORM"] == "darwin":
        env.Append(CXXFLAGS=" -undefined dynamic_lookup")
        env.Append(SHLINKFLAGS=" -undefined dynamic_lookup")
        env.Append(LDMODULEFLAGS=" -undefined dynamic_lookup")
        if not env.GetOption("clean"):
            # remove /usr/lib if present in LIBPATH.
            # there should maybe be some testing here to make sure 
            # it doesn't blow in our face.
            try: env["LIBPATH"].remove('/usr/lib')
            except: pass
        env['SHLINKFLAGS'] = env['LDMODULEFLAGS']
        env['SHLINKCOM']   = env['LDMODULECOM']
        env["SHLIBSUFFIX"] = ".so"
    elif env["PLATFORM"].startswith("win"):
        env["SHLIBSUFFIX"] = ".pyd"

    # Set flags for swig command
    env["SWIGFLAGS"] = "-python -c++ -shadow -O ".split() +\
                       ["-DPACKAGE_VERSION=%s"%env["PACKAGE_VERSION"]]

    # Attach a swig scanner to the environment
    env.Prepend(SCANNERS=SwigScanner)

    return env

    
def conf_message(conf, message, to_screen=False, somthingelse=False):
    "Help function to write to log file and screen" 
    conf.logstream.write(message)
    if to_screen:
        print message

# Taken from http://ivory.idyll.org/blog/mar-07/replacing-commands-with-subprocess
def get_status_output(cmd, input=None, cwd=None, env=None):
    from subprocess import Popen, PIPE, STDOUT
    pipe = Popen(cmd, shell=True, cwd=cwd, env=env, stdout=PIPE, stderr=STDOUT)
    (output, errout) = pipe.communicate(input=input)
    assert not errout
    status = pipe.returncode
    return (status, output)

def get_package_dir(package, env, header=None, library=None):
    """Return True together with the directory if a header from a package is found
False with a message otherwise"""
    assert(isinstance(package, str))
    assert(isinstance(env, SCons.Script.Environment))
    assert(header or library)

    # Check user specified directories
    if env is not None and \
           env.get("with_%s_dir" % package.lower(), None):
        package_dir = env["with_%s_dir" % package.lower()]
    elif os.environ.has_key('%s_DIR' % package.upper()):
        package_dir = os.environ["%s_DIR" % package.upper()]
    else:
        package_dir = env["DEFAULT_LIB_DIR"]
    
    is_found = False

    # Check for the existance of the passed header file
    if header:
        message = "Could not find header for %s\nSearched directory: '%s'%s[include] for %s\n"%\
                  (package, package_dir, os.path.sep, header)
        for inc_dir in ["", "include"]:
            if os.path.isfile(os.path.join(package_dir, inc_dir, header)):
                package_dir = os.path.join(package_dir, inc_dir)
                is_found = True
                break
    else:
        message = "Could not find library for %s\nSearched directory: '%s'%s[lib] for %s\n"%\
                  (package, package_dir, os.path.sep, library)
        for lib_dir in ["", "lib"]:
            for prefix in env["LIBPREFIXES"]:
                for suffix in env["LIBSUFFIXES"]:
                    lib_file = env.subst(prefix)+library+env.subst(suffix)
                    if os.path.isfile(os.path.join(package_dir, lib_dir, lib_file)):
                        package_dir = os.path.join(package_dir, lib_dir)
                        is_found = True
                        break
        

    if is_found:
        return True, package_dir
    
    return False, message

from packages import *
from scons_tools import is_darwin, is_windows, SwigScanner
