__author__ = "Johan Hake (hake.dev@gmail.com)"
__date__ = "2010-05-25 -- 2010-09-02"
__copyright__ = "Copyright (C) 2010 " + __author__
__license__  = "GNU LGPL Version 3.0 or later"

__all__ = ['get_default_prefixes',
           'init_environment', 'handle_scons_options', 'handle_compile_flags',
           'is_windows', 'is_darwin', 'generate_pkg_config_file',
           'generate_runtime_conf_file', 'configure_swig_env', 'configure_env',
           'print_exit_message', 'SwigScanner']

import os, sys, re, SCons, compileall, atexit
from SCons.Script import *
from distutils  import sysconfig

def init_environment(name, version="", description="",
                     default_prefix=None,
                     default_python_prefix=None,
                     default_lib_dir=None):

    # Init the SCons environment from the os environment
    env = Environment(ENV=os.environ,
                      BUILDERS = {"ByteCompile"   : ByteCompile,
                                  "PkgConfigFile" : PkgConfigFile,
                                  "ConfFile"      : ConfFile,
                                  "Textfile"      : Textfile})
    
    env['PACKAGE']               = name
    env['PACKAGE_VERSION']       = version
    env['PACKAGE_DSCR']          = description
    env['END_MESSAGE']           = ""
    env['CONFIG_FILE_GENERATED'] = False
    env['INSTALL_DIRS']          = {}
    env['PACKAGE_LIBS']          = {}
    env['PACKAGE_DEPS']          = {"DEPENDENCIES":[], "LIBS":[], "LIBPATH":[],
                                    "CPPPATH":[]}

    # Platform specific initializations
    if is_windows(env):
        env['DEFAULT_PREFIX'] = default_prefix or r"C:\local"
        env['DEFAULT_PYTHON_PREFIX'] = default_python_prefix or "$prefix\Lib\site-packages"
        env['DEFAULT_LIB_DIRo'] = default_lib_dir or r"C:\local"
    elif is_darwin(env):
        env['DEFAULT_PREFIX'] = default_prefix or "/sw"
        env['DEFAULT_PYTHON_PREFIX'] = default_python_prefix or \
                                       os.path.join("$prefix/lib/python%s/site-packages"%
                                             sysconfig.get_python_version())
        env['DEFAULT_LIB_DIR'] = default_lib_dir or "/sw"
    else:
        env['DEFAULT_PREFIX'] = default_prefix or "/usr/local"
        env['DEFAULT_PYTHON_PREFIX'] = default_python_prefix or \
                                       sysconfig.get_python_lib(prefix="$prefix",
                                                                plat_specific=True)
        env['DEFAULT_LIB_DIR'] = default_lib_dir or "/usr"

    env.Append(CPPDEFINES={"PACKAGE_VERSION":version})

    # Return an updated SCons environment
    return env

def print_exit_message(env):
    # Define what happens at exit
    def when_finished():
        if env.GetOption('help'):
            return 
        if env.GetOption("clean"):
            # Clean any cached files
            if env["clean_cache"]:
                print "Removed cache"
                for f in [env["cache_file"], ".sconsign.dblite"]:
                    os.unlink(f)
                for f in Glob(os.path.join(".sconf_temp","*")):
                    os.unlink(str(f))
            return
        if 'install' in SCons.Script.COMMAND_LINE_TARGETS:
            env['END_MESSAGE'] +="""
---------------------------------------------------------
Installation of %(PACKAGE)s finished.
"""%env
            if env['CONFIG_FILE_GENERATED']:
                cmd = "%s.bat"%env["PACKAGE"] if is_windows(env) else \
                          "source %s.conf"%env["PACKAGE"]
                env['END_MESSAGE'] += """
To setup a runtime environment execute:

   %s
"""%cmd

                env['END_MESSAGE'] +="""
---------------------------------------------------------"""
        else:
            env['END_MESSAGE'] += """
---------------------------------------------------------
Compilation of %(PACKAGE)s finished. Now execute

    scons install

to install %(PACKAGE)s on your system. Note that you may
need to be root in order to install. To specify
an alternative installation directory, execute

    scons install prefix=<path>

---------------------------------------------------------
"""%env

        from SCons.Script import GetBuildFailures
        build_failures = GetBuildFailures()
        if build_failures:
            for bf in build_failures:
                print "%s failed: %s" % (bf.node, bf.errstr)
            return
        print env['END_MESSAGE']

    # Do the exit stuff
    atexit.register(when_finished)
    

def get_default_prefixes(env, posix_default="/usr/local",
                         darwin_default="/sw",
                         win_default=r"c:\local"):
    "Return the default prefix and python dir"
    assert(isinstance(env, SCons.Script.Environment))

    return default_prefix, default_python_dir

def is_darwin(env):
    ""
    assert(isinstance(env, SCons.Script.Environment))
    return env["PLATFORM"] == "darwin"

def is_windows(env):
    assert(isinstance(env, SCons.Script.Environment))
    return env["PLATFORM"].startswith("win")    

def handle_scons_options(env, options, cache_file="options.cache"):
    "Update environment with the user specified options"
    assert(isinstance(env, SCons.Script.Environment))
    assert(isinstance(options, list))

    # Default options
    default_options = [
        BoolVariable("enable_debug", "Build with debug information", 1),
        BoolVariable("enable_optimize", "Compile with optimization", 0),
        BoolVariable("cache_options", "Cache command-line options for later invocations", 1),
        BoolVariable("clean_cache", "Remove the cached options and configuration files. Used during cleaning", 0),
        ("custom_CXX_flags", "Customize compilation of C++ code", ""),
        ("custom_link_flags", "Customize linking of C++ code", "")
        ]

    # Append default options
    options.extend(default_options)

    # Set the options using any cached options
    opts = SCons.Variables.Variables(cache_file, args=SCons.Script.ARGUMENTS.copy())
    opts.AddVariables(*options)
    option_length = reduce(max, (len(s) for s in opts.keys()), 0)
    def my_format(env, opt, help, default, actual, l):
        fmt = "%s %s\n%sactual='%s'\n\n"
        return fmt % ((opt+":").ljust(option_length+2), help, " "*(option_length+3), actual)
    opts.FormatVariableHelpText = my_format

    opts.Update(env)

    if opts.UnknownVariables():
        for key, value in opts.UnknownVariables().iteritems():
            print "unknown option:  %s=%s" % (key, value)
        Exit(1)
            
    # Make any local prefix dir absolute
    env["prefix"] = os.path.abspath(env["prefix"])

    # Pick out installation directories
    for install_dir in ["prefix", "pkgconfig_dir", "include_dir", "lib_dir",
                        "python_module_dir", "bin_dir"]:
        if env.has_key(install_dir):
            env["INSTALL_DIRS"][install_dir] = env[install_dir]

    # Set alias for install
    env.Alias("install", env["INSTALL_DIRS"].values())

    # Now generate the help text
    env.Help(opts.GenerateHelpText(env))

    # Save options to cach file
    if env["cache_options"]:
        cache_options = env["cache_options"]
        clean_cache = env["clean_cache"]
        del env["cache_options"] # Don't store this value
        del env["clean_cache"]
        
        opts.Save(cache_file, env)
        
        # Restore options
        env["cache_options"] = cache_options
        env["clean_cache"] = clean_cache

    # Store the cache file
    env["cache_file"] = cache_file
    
    # If not cleaning
    if not env.GetOption("clean"):
        # Notify the user about that options from scons/options.cache are being used:
        if os.path.isfile(cache_file) and os.path.getsize(cache_file):
            print "Using cached options from '%s'"%cache_file

def handle_compile_flags(env, default_flags="",
                         default_link_flags="",
                         default_debug_flags="-g",
                         default_optimization_flags="-O2"):
    # Set default and user defined link and cpp flags
    env['CXXFLAGS'] = os.environ.get("CXXFLAGS", default_flags)
    env.Append(CXXFLAGS=env["custom_CXX_flags"])
    
    env["LINKFLAGS"] = os.environ.get("LINKFLAGS", default_link_flags)
    env.Append(LINKFLAGS=env["custom_link_flags"])

    if env["enable_debug"] and not env["enable_optimize"]:
        env.Append(CXXFLAGS=" %s"%default_debug_flags)
    if env["enable_optimize"]:
        env.Append(CXXFLAGS=" %s"%default_optimization_flags)

def compile_python_modules(target, source, env):
    "Byte compiles python modules"
    compileall.compile_file(str(source[0]), quiet=True)

def compile_python_modules_output(target, source, env):
    return "Byte compiling %s"%str(source[0])

def swig_scanner(node, env, path):
    """SCons scanner hook to scan swig files for dependencies.
    @return: list of files depended on.
    """
    include_pattern = re.compile(r"%include\s+(\S+)")
    import_pattern  = re.compile(r"%import\s+(\S+)")

    def recurse(path, search_path):
        """Scan recursively."""
        f = open(path)
        try: contents = f.read()
        finally: f.close()

        found = []
        for m in include_pattern.finditer(contents):
            inc_fpath = m.group(1)
            # Strip off quotes
            inc_fpath = inc_fpath.strip("'").strip('"')
            # Strip off site include marks
            inc_fpath = inc_fpath.strip('<').strip('>')
            real_fpath = os.path.realpath(inc_fpath)
            if os.path.dirname(real_fpath) != os.path.dirname(path):
                # Not in same directory as original swig file
                if os.path.isfile(real_fpath):
                    found.append(real_fpath)
                    continue
            
            # Look for unqualified filename on path
            for dpath in search_path:
                abs_path = os.path.join(dpath, inc_fpath)
                if os.path.isfile(abs_path):
                    found.append(abs_path)
                    break
        
        for m in import_pattern.finditer(contents):
            inc_fpath = m.group(1)
            # Strip off quotes
            inc_fpath = inc_fpath.strip("'").strip('"')
            # Strip off site include marks
            inc_fpath = inc_fpath.strip('<').strip('>')
            real_fpath = os.path.realpath(inc_fpath)
            if os.path.dirname(real_fpath) != os.path.dirname(path):
                # Not in same directory as original swig file
                if os.path.isfile(real_fpath):
                    found.append(real_fpath)
                    continue
            
            # Look for unqualified filename on path
            for dpath in search_path:
                abs_path = os.path.join(dpath, inc_fpath)
                if os.path.isfile(abs_path):
                    found.append(abs_path)
                    break
        
        for f in [f for f in found if os.path.splitext(f)[1] == ".i"]:
            found += recurse(f, search_path)
        return found
    
    fpath = node.srcnode().path
    search_path = [os.path.abspath(d) for d in re.findall(r"-I(\S+)", " ".join(env["SWIGFLAGS"]))]
    search_path.insert(0, os.path.abspath(os.path.dirname(fpath)))
    r = recurse(fpath, search_path)
    return r

SwigScanner = SCons.Scanner.Scanner(function=swig_scanner, skeys=[".i"])

def generate_pkg_config(target, source, env):
    file = open(str(target[0]), "w")
    source = eval(source[0].get_contents())
    cflags = [env.subst(env["include_dir"])]
    libs = [env['PACKAGE'].lower()]
    lib_dirs = [env.subst(env["lib_dir"])]
    file.write("Name: %s\n" % env['PACKAGE'].capitalize())
    if env["PACKAGE_VERSION"]:
        file.write("Version: %s\n" % env["PACKAGE_VERSION"])
    if env["PACKAGE_DSCR"]:
        file.write("Description: %s\n" % env["PACKAGE_DSCR"])
    if not source["inline_external_dependencies"]:
        file.write("Requires:  %s\n"%(" ".join(env["PACKAGE_DEPS"]["DEPENDENCIES"])))
    else:
        cflags.extend(env['PACKAGE_DEPS']['CPPPATH'])
        lib_dirs.extend(env['PACKAGE_DEPS']['LIBPATH'])
        libs.extend(env['PACKAGE_DEPS']['LIBS'])

    file.write("Cflags: %s\n" % (" ".join("-I%s"%s for s in cflags)))
    file.write("Libs: %s %s\n" % (" ".join("-L%s"%s for s in lib_dirs),\
                                  " ".join("-l%s"%s for s in libs)))
    file.close()
    
def generate_file_output(target, source, env):
    return "Generating '%s'"%str(target[0])

def generate_conf_file(target, source, env):
    f = open(str(target[0]), 'w')
    bin_dir = env.subst(env.get('bin_dir', ''))
    lib_dir = env.subst(env.get('lib_dir', ''))
    pkgconfig_dir = env.subst(env.get('pkgconfig_dir', ''))
    python_module_dir = env.subst(env.get('python_module_dir', ''))
    sep = os.pathsep
    if is_windows(env):
        if bind_dir or lib_dir:
            f.write('set PATH=%s\n' % sep.join([bin_dir, lib_dir, '%PATH%']))
        if python_module_dir:
            f.write('set PYTHONPATH=%s\n' % \
                    sep.join([python_module_dir, '%PYTHONPATH%']))
        if pkgconfig_dir:
            f.write('set PKG_CONFIG_PATH=%s\n' % \
                    sep.join([pkgconfig_dir, '%PKG_CONFIG_PATH%']))
        f.close()
    else:
        if lib_dir:
            if is_darwin(env):
                f.write('export DYLD_LIBRARY_PATH=%s\n' % \
                        sep.join([lib_dir, '$DYLD_LIBRARY_PATH']))
            else:
                f.write('export LD_LIBRARY_PATH=%s\n' % \
                        sep.join([lib_dir, '$LD_LIBRARY_PATH']))
        if bin_dir:
            f.write('export PATH=%s\n' % sep.join([bin_dir, '$PATH']))
        if python_module_dir:
            f.write('export PYTHONPATH=%s\n' % \
                    sep.join([python_module_dir, '$PYTHONPATH']))
        if pkgconfig_dir:
            f.write('export PKG_CONFIG_PATH=%s\n' % \
                    sep.join([pkgconfig_dir, '$PKG_CONFIG_PATH']))
        f.close()

def generate_text_file(target, source, env):
    f = open(str(target[0]), 'w')
    f.write("\n".join(eval(source[0].get_contents())))
    f.close()


ByteCompile = SCons.Script.Builder(action=Action(compile_python_modules,
                                   strfunction=compile_python_modules_output),
                                   suffix=".pyc", src_suffix=".py")

PkgConfigFile = SCons.Script.Builder(action=Action(generate_pkg_config,
                                     strfunction=generate_file_output))

ConfFile = SCons.Script.Builder(action=Action(generate_conf_file,
                                strfunction=generate_file_output))

Textfile = SCons.Script.Builder(action=Action(generate_text_file,
                                strfunction=generate_file_output))

def generate_pkg_config_file(env, inline_external_dependencies=True):
    "Return a pkg-config file and a conf file"
    assert(isinstance(env, SCons.Script.Environment))

    source = env["INSTALL_DIRS"].copy()
    source["prefix"]       = env["prefix"]
    source["name"]         = env["PACKAGE"]
    source["dscr"]         = env["PACKAGE_DSCR"]
    source["version"]      = env["PACKAGE_VERSION"]
    source["package_deps"] = env["PACKAGE_DEPS"]
    source["inline_external_dependencies"] = inline_external_dependencies
    source =  Value(source)

    pkg_config = env.PkgConfigFile("%s.pc"%env['PACKAGE'], source)

    # Setup default build target
    env.Default(pkg_config)

    # Setup install targets
    env.Install(env["pkgconfig_dir"], pkg_config)

def generate_runtime_conf_file(env):
    "Generate a conf file that can be 'sourced to set the correct environment'"
    source = Value(env["INSTALL_DIRS"].copy())
    
    conf_suffix = ".bat" if is_windows(env) else ".conf"
    conf_file   = env.ConfFile(env['PACKAGE']+conf_suffix, source)
    env['CONFIG_FILE_GENERATED'] = True

    # Setup default build target
    env.Default(conf_file)

# Put here to avoid circular dependency
from scons_configure import *
