// Copyright (C) 2010 Johan Hake.
//
// First added:  2010-01-21
// Last changed: 2010-06-01
//
// Unit test for the MarkovModel class

#include <iomanip>

#include <gillstep.h>
#include <gillstep/unittest.h>

#include "Constant.h"

using namespace gillstep;

typedef std::vector<MarkovModel*>::iterator mm_iter;

// Create local gillstepper
GillespieStepper gillstepper;
    
class GillstepInterFace : public CppUnit::TestFixture
{
  CPPUNIT_TEST_SUITE(GillstepInterFace);
  CPPUNIT_TEST(test_create_and_destroy);
  CPPUNIT_TEST(test_mm_reset_and_step);
  CPPUNIT_TEST_SUITE_END();

public:

  void test_create_and_destroy()
  {
    // Create Constant MarkovModel
    std::vector<MarkovModel*> mm_vec;
    mm_vec.push_back(new Constant(gillstepper));
    mm_vec.push_back(new Constant(gillstepper));
    mm_vec.push_back(new Constant(gillstepper));
    mm_vec.push_back(new Constant(gillstepper));
    mm_vec.push_back(new Constant(gillstepper));

    CPPUNIT_ASSERT(gillstepper.num_markov_models() == mm_vec.size());

    delete mm_vec[3];
    delete mm_vec[4];
    mm_vec.pop_back();
    mm_vec.pop_back();
    
    CPPUNIT_ASSERT(gillstepper.num_markov_models() == mm_vec.size());
    
    // Clean up
    for (mm_iter it = mm_vec.begin(); it != mm_vec.end(); it++)
      delete *it;
  }

  void test_mm_reset_and_step()
  {
    // Create Constant MarkovModel
    std::vector<MarkovModel*> mm_vec;
    mm_vec.push_back(new Constant(gillstepper));
    mm_vec.push_back(new Constant(gillstepper));
    mm_vec.push_back(new Constant(gillstepper));

    // Test reseting using deterministic instead of random initialization
    gillstepper.set_deterministic(true, 0.05);
    gillstepper.reset();

    for (mm_iter it = mm_vec.begin(); it != mm_vec.end(); it++)
      CPPUNIT_ASSERT((*it)->state() == 0);
      
    gillstepper.set_deterministic(true, 0.15);
    gillstepper.reset();

    for (mm_iter it = mm_vec.begin(); it != mm_vec.end(); it++)
      CPPUNIT_ASSERT((*it)->state() == 1);
      
    gillstepper.set_deterministic(true, 0.25);
    gillstepper.reset();

    for (mm_iter it = mm_vec.begin(); it != mm_vec.end(); it++)
      CPPUNIT_ASSERT((*it)->state() == 2);
      
    gillstepper.set_deterministic(true, 0.8);
    gillstepper.reset();

    for (mm_iter it = mm_vec.begin(); it != mm_vec.end(); it++)
      CPPUNIT_ASSERT((*it)->state() == 3);
      
    // Check sum of rates
    CPPUNIT_ASSERT(gillstepper.sum_rates() == 3.0);

    // Step a markov model, as all are in state 3 this should return true
    CPPUNIT_ASSERT(gillstepper.step_markov_model());
    CPPUNIT_ASSERT(mm_vec[2]->state() == 2);
    CPPUNIT_ASSERT(gillstepper.sum_rates() == 5.0);

    // Change deterministic step to 0.001
    gillstepper.set_deterministic(true, 0.001);

    // Step a markov model, as the first Markov model is in state 3 this should
    // return true
    CPPUNIT_ASSERT(gillstepper.step_markov_model());
    CPPUNIT_ASSERT(mm_vec[0]->state() == 2);
    CPPUNIT_ASSERT(gillstepper.sum_rates() == 7.0);

    // Check time to next transition
    gillstepper.set_deterministic(true, 0.5);
    gillstepper.new_dimless_exit_time();

    CPPUNIT_ASSERT(fabs(gillstepper.next_transition() - 0.099021025794277892) < 1e-16);
    
    // Update time 
    CPPUNIT_ASSERT(gillstepper.step_dt_unless_transition(0.05) == 0.0);
    CPPUNIT_ASSERT(fabs(gillstepper.time()-0.05) < 1e-16);
    CPPUNIT_ASSERT(fabs(gillstepper.next_transition() - 0.049021025794277892) < 1e-16);
    
    // Update time 
    CPPUNIT_ASSERT(fabs(gillstepper.step_dt_unless_transition(0.5) - 0.45097897420572208) < 1e-16);
    CPPUNIT_ASSERT(fabs(gillstepper.time()-0.099021025794277892) < 1e-16);
    CPPUNIT_ASSERT(fabs(gillstepper.next_transition() - 0.0770163533955494) < 1e-16);
    
    gillstepper.step_dt(4.9009789742057217);
    
    CPPUNIT_ASSERT(fabs(gillstepper.time() - 5.0) < 1e-16);
    // Clean up
    for (mm_iter it = mm_vec.begin(); it != mm_vec.end(); it++)
      delete *it;
    
  }

};

CPPUNIT_TEST_SUITE_REGISTRATION(GillstepInterFace);

int main()
{
  GILLSTEP_TEST;
}
