"""Unit tests for the compiled Markov models"""

__author__ = "Johan Hake (hake.dev@gmail.com)"
__date__ = "2010-06-01 -- 2010-11-24"
__copyright__ = "Copyright (C) 2010 " + __author__
__license__  = "GNU LGPL Version 3.0 or later"

import unittest
from gillstep import *

gillstepper = GillespieStepper()

constant_model = ModelEditor("Constant")
constant_model.set_Q10(2, 37)

constant_model.rates[0,1] = 2
constant_model.rates[1,2] = 2
constant_model.rates[2,3] = 2

constant_model.rates[1,0] = 1
constant_model.rates[2,1] = 1
constant_model.rates[3,2] = 1

# State values
constant_model.state_values[3] = 1

# Simple variable dependent model
simple_model = ModelEditor("Simple")
simple_model.set_variables("V")

simple_model.rates[0,1] = V**2 
simple_model.rates[1,0] = 1

simple_model.state_values[1] = 1

class ConstantInterFace(unittest.TestCase):

    def test_create_and_destroy(self):
        # Create Constant MarkovModel
        mms = jit(constant_model, gillstepper, 5)
        self.assertEqual(gillstepper.num_markov_models(), len(mms))

        mms.pop()
        mms.pop()

        self.assertEqual(gillstepper.num_markov_models(), len(mms))
        
    def test_mm_reset_and_step(self):
        # Create Constant MarkovModel
        mms = jit(constant_model, gillstepper, 3)
        
        # Test reseting using deterministic instead of random initialization
        gillstepper.set_deterministic(True, 0.05)
        gillstepper.reset()

        for mm in mms:
            self.assertEqual(mm.state(), 0)
        
        gillstepper.set_deterministic(True, 0.15)
        gillstepper.reset()

        for mm in mms:
            self.assertEqual(mm.state(), 1)
        
        gillstepper.set_deterministic(True, 0.25)
        gillstepper.reset()

        for mm in mms:
            self.assertEqual(mm.state(), 2)
        
        gillstepper.set_deterministic(True, 0.8)
        gillstepper.reset()

        for mm in mms:
            self.assertEqual(mm.state(), 3)

        # Check sum of rates
        self.assertEqual(gillstepper.sum_rates(), 3.0)
        
        # Step last markov model, as all are in state 3 this should return True
        self.assertTrue(gillstepper.step_markov_model())
        self.assertEqual(mms[-1].state(), 2)
        self.assertEqual(gillstepper.sum_rates(), 5.0)

        # Change deterministic ste to 0.001
        gillstepper.set_deterministic(True, 0.001)

        # Step first markov model, as it is in state 3 this should return True
        self.assertTrue(gillstepper.step_markov_model())
        self.assertEqual(mms[0].state(), 2)
        self.assertEqual(gillstepper.sum_rates(), 7.0)

        # Check time to next transition
        gillstepper.set_deterministic(True, 0.5)
        gillstepper.new_dimless_exit_time()

        self.assertAlmostEqual(gillstepper.next_transition(), 0.099021025794277892)
    
        # Update time 
        self.assertAlmostEqual(gillstepper.step_dt_unless_transition(0.05), 0.0)
        self.assertAlmostEqual(gillstepper.time(), 0.05)
        self.assertAlmostEqual(gillstepper.next_transition(), 0.049021025794277892)
    
        # Update time 
        self.assertAlmostEqual(gillstepper.step_until_next_transition(), 0.049021025794277892)
        self.assertAlmostEqual(gillstepper.time(), 0.099021025794277892)
        self.assertAlmostEqual(gillstepper.next_transition(), 0.0770163533955494)
        gillstepper.step_dt(4.9009789742057217)
        self.assertAlmostEqual(gillstepper.time(), 5.0)

    def test_substepping(self):
        # Create simple MarkovModel
        mm = jit(simple_model, gillstepper)

        # Send initial values of variables
        mm.receive("V")(1.0)
        gillstepper.set_deterministic(True, 0.4)
        gillstepper.reset()
        mm.receive("V")(100.0)
        substep = 0
        while substep < 1.0:
            oldsubstep = substep
            substep = mm.substep_size(oldsubstep)
            rate0 = mm.evaluate_substep_rate(0, oldsubstep, oldsubstep)
            rate1 = mm.evaluate_substep_rate(0, oldsubstep, substep)
            self.assertTrue(abs(rate1/rate0-1) - gillstepper.get_max_fractional_rate_change() <= 1e-15)

    def test_error_convergence(self):
        # Create simple MarkovModel
        from math import log
        mm = jit(simple_model, gillstepper)
        
        gillstepper.set_deterministic(True, 0.4)

        def integrated_rate(V0, V1, macro_timestep, timestep):
            "Integrate a rate symbolicly"
            from sympy import Symbol, integrate
            t = Symbol("t")
            V_of_t = V0 + (V1-V0)/macro_timestep*t
            rate = simple_model.rates[0,1].subs(V, V_of_t)
            return integrate(rate, (t, 0.0, timestep))
        
        timestep = 0.1
        V0 = 1.0
        V1 = 10.0
        prev_error = 1.0
        for rate_change in [0.1, 0.01, 0.001, 0.0001]:
            # Send initial values of variables
            mm.receive("V")(V0)
            gillstepper.reset()
            mm.receive("V")(V1)
            gillstepper.set_max_fractional_rate_change(rate_change)
            this_error = abs(1-(-log(0.4))/integrated_rate(V0, V1, timestep,
                            timestep-gillstepper.step_dt_unless_transition(
                                                           timestep, True)))
            self.assertTrue(prev_error>this_error)
            prev_error = this_error

        V0 = 10.0
        V1 = 1.0
        prev_error = 1.0
        for rate_change in [0.1, 0.01, 0.001, 0.0001]:
            # Send initial values of variables
            mm.receive("V")(V0)
            gillstepper.set_deterministic(True, 0.0000001)
            gillstepper.reset()
            gillstepper.set_deterministic(True, 0.1)
            gillstepper.new_dimless_exit_time()
            mm.receive("V")(V1)
            gillstepper.set_max_fractional_rate_change(rate_change)
            this_error = abs(1-(-log(0.1))/integrated_rate(V0, V1, timestep,
                            timestep-gillstepper.step_dt_unless_transition(
                                                           timestep, True)))
            self.assertTrue(prev_error>this_error)
            prev_error = this_error
        

if __name__ == "__main__":
    unittest.main()
