// Copyright (C) 2010 Johan Hake.
//
// First added:  2010-01-21
// Last changed: 2010-09-02
//
// Unit test for the MarkovModel class

#include <gillstep.h>
#include <gillstep/unittest.h>

#include "Constant.h"

using namespace gillstep;

// Create local stepper
GillespieStepper gillstepper;

class ConstantInterFace : public CppUnit::TestFixture
{
  CPPUNIT_TEST_SUITE(ConstantInterFace);
  CPPUNIT_TEST(test_basic);
  CPPUNIT_TEST(test_eval_rate);
  CPPUNIT_TEST(test_state_transition);
  CPPUNIT_TEST(test_temp_setting);
  CPPUNIT_TEST_SUITE_END();

public:

  void test_basic()
  {
    // Create Constant MarkovModel
    Constant mm(gillstepper);
    CPPUNIT_ASSERT(mm.max_transitions() == 2);
    CPPUNIT_ASSERT(mm.num_states() == 4);
    CPPUNIT_ASSERT(mm.num_submodels() == 0);
  }

  void test_eval_rate()
  {
    // Create Constant MarkovModel
    Constant mm(gillstepper);
    gillstep::uint i, j;
    i=0; j=1; CPPUNIT_ASSERT(mm.evaluate_rate(i,j) == 2.0);
    i=1; j=2; CPPUNIT_ASSERT(mm.evaluate_rate(i,j) == 2.0);
    i=2; j=3; CPPUNIT_ASSERT(mm.evaluate_rate(i,j) == 2.0);

    i=1; j=2; CPPUNIT_ASSERT(mm.evaluate_rate(i,j) == 2.0);
    i=2; j=3; CPPUNIT_ASSERT(mm.evaluate_rate(i,j) == 2.0);
    i=1; j=0; CPPUNIT_ASSERT(mm.evaluate_rate(i,j) == 1.0);
    i=2; j=1; CPPUNIT_ASSERT(mm.evaluate_rate(i,j) == 1.0);
    i=3; j=2; CPPUNIT_ASSERT(mm.evaluate_rate(i,j) == 1.0);

    i=0; j=0; CPPUNIT_ASSERT(mm.evaluate_rate(i,j) == 0.0);
    i=0; j=2; CPPUNIT_ASSERT(mm.evaluate_rate(i,j) == 0.0);
    i=0; j=3; CPPUNIT_ASSERT(mm.evaluate_rate(i,j) == 0.0);

    i=1; j=1; CPPUNIT_ASSERT(mm.evaluate_rate(i,j) == 0.0);
    i=1; j=3; CPPUNIT_ASSERT(mm.evaluate_rate(i,j) == 0.0);

    i=2; j=0; CPPUNIT_ASSERT(mm.evaluate_rate(i,j) == 0.0);
    i=2; j=2; CPPUNIT_ASSERT(mm.evaluate_rate(i,j) == 0.0);

    i=3; j=0; CPPUNIT_ASSERT(mm.evaluate_rate(i,j) == 0.0);
    i=3; j=1; CPPUNIT_ASSERT(mm.evaluate_rate(i,j) == 0.0);
    i=3; j=3; CPPUNIT_ASSERT(mm.evaluate_rate(i,j) == 0.0);

  }
   
  void test_state_transition()
  {
    // Create Constant MarkovModel
    Constant mm(gillstepper);
    
    // State 0
    CPPUNIT_ASSERT(mm.state() == 0);
    CPPUNIT_ASSERT(mm.value() == 0.0);
    CPPUNIT_ASSERT(mm.evaluate_rate(0) == 2.0);
    CPPUNIT_ASSERT(mm.change_state(0) == false);

    // State 1
    CPPUNIT_ASSERT(mm.state() == 1);
    CPPUNIT_ASSERT(mm.value() == 0.0);
    CPPUNIT_ASSERT(mm.evaluate_rate(0) == 1.0);
    CPPUNIT_ASSERT(mm.evaluate_rate(1) == 2.0);
    CPPUNIT_ASSERT(mm.change_state(1) == false);

    // State 2
    CPPUNIT_ASSERT(mm.state() == 2);
    CPPUNIT_ASSERT(mm.value() == 0.0);
    CPPUNIT_ASSERT(mm.evaluate_rate(0) == 1.0);
    CPPUNIT_ASSERT(mm.evaluate_rate(1) == 2.0);
    CPPUNIT_ASSERT(mm.change_state(1) == true);

    // State 3
    CPPUNIT_ASSERT(mm.state() == 3);
    CPPUNIT_ASSERT(mm.value() == 1.0);
    CPPUNIT_ASSERT(mm.evaluate_rate(0) == 1.0);
    CPPUNIT_ASSERT(mm.change_state(0) == true);

    // State 2
    CPPUNIT_ASSERT(mm.state() == 2);
    CPPUNIT_ASSERT(mm.value() == 0.0);
    CPPUNIT_ASSERT(mm.evaluate_rate(0) == 1.0);
    CPPUNIT_ASSERT(mm.evaluate_rate(1) == 2.0);
    CPPUNIT_ASSERT(mm.change_state(0) == false);

    // State 1
    CPPUNIT_ASSERT(mm.state() == 1);
    CPPUNIT_ASSERT(mm.value() == 0.0);
    CPPUNIT_ASSERT(mm.evaluate_rate(0) == 1.0);
    CPPUNIT_ASSERT(mm.evaluate_rate(1) == 2.0);
    CPPUNIT_ASSERT(mm.change_state(0) == false);

    // State 0
    CPPUNIT_ASSERT(mm.state() == 0);
    CPPUNIT_ASSERT(mm.value() == 0.0);
    CPPUNIT_ASSERT(mm.evaluate_rate(0) == 2.0);

  }

  void test_temp_setting()
  {
    // Create Constant MarkovModel
    Constant mm(gillstepper);
    gillstep::uint i, j;
    
    i=0;j=1;CPPUNIT_ASSERT(mm.evaluate_rate(i,j) == 2.0);
    i=3;j=2;CPPUNIT_ASSERT(mm.evaluate_rate(i,j) == 1.0);

    set_T(27);
    mm.update_temp_scale();

    i=0;j=1;CPPUNIT_ASSERT(mm.evaluate_rate(i,j) == 1.0);
    i=3;j=2;CPPUNIT_ASSERT(mm.evaluate_rate(i,j) == 0.5);

    set_T(47);
    mm.update_temp_scale();

    i=0;j=1;CPPUNIT_ASSERT(mm.evaluate_rate(i,j) == 4.0);
    i=3;j=2;CPPUNIT_ASSERT(mm.evaluate_rate(i,j) == 2.0);

    mm.set_parameter("T_ref", 47.0);
    i=0;j=1;CPPUNIT_ASSERT(mm.evaluate_rate(i,j) == 2.0);
    i=3;j=2;CPPUNIT_ASSERT(mm.evaluate_rate(i,j) == 1.0);

    mm.set_parameter("T_ref", 37.0);
    mm.set_parameter("Q10", 1.0);

    i=0;j=1;CPPUNIT_ASSERT(mm.evaluate_rate(i,j) == 2.0);
    i=3;j=2;CPPUNIT_ASSERT(mm.evaluate_rate(i,j) == 1.0);

  }  

};
CPPUNIT_TEST_SUITE_REGISTRATION(ConstantInterFace);

int main()
{
  GILLSTEP_TEST;
}
