"""Unit tests for the compiled Markov models"""

__author__ = "Johan Hake (hake.dev@gmail.com)"
__date__ = "2010-06-01 -- 2011-01-12"
__copyright__ = "Copyright (C) 2010 " + __author__
__license__  = "GNU LGPL Version 3.0 or later"

import unittest
from gillstep import *

gillstepper = GillespieStepper()
constant_model = ModelEditor("Constant")
constant_model.set_Q10(2, 37)

constant_model.rates[0,1] = 2
constant_model.rates[1,2] = 2
constant_model.rates[2,3] = 2

constant_model.rates[1,0] = 1
constant_model.rates[2,1] = 1
constant_model.rates[3,2] = 1

# State values
constant_model.state_values[3] = 1

simple_model = ModelEditor("simple")

# Declarations
simple_model.set_parameters(alpha_0 = 1.0,
                            alpha_1 = 2.0,
                            beta_0  = 3.0,
                            beta_1  = 4.0,
                            gamma_0 = 5.0,
                            a       = 6.0)

simple_model.set_variables("V", "dyadic_Ca")

simple_model.set_Q10(1.8, 37)

# Rate functions
alpha = exp(-(20-V/(R*T))/10)
beta  = exp(-(V-35)/3)
gamma = gamma_0*dyadic_Ca

simple_model.rates[0,1] = 2*alpha_0*alpha
simple_model.rates[1,2] = alpha_1*alpha
simple_model.rates[1,0] = beta_0*beta
simple_model.rates[2,1] = 2*beta_1*beta
simple_model.rates[3,2] = gamma
simple_model.rates[2,3] = a**2

# State values
simple_model.state_values[3] = 1

# A global value
received_value = 1.0
        
class ConstantInterFace(unittest.TestCase):

    def test_basic(self):
        # Create Constant MarkovModel
        mm = jit(constant_model, gillstepper)
        self.assertEqual(mm.max_transitions(), 2)
        self.assertEqual(mm.num_states(), 4)
        self.assertEqual(mm.num_submodels(), 0)
        
    def test_eval_rate(self):
        # Create Constant MarkovModel
        mm = jit(constant_model, gillstepper)
        self.assertEqual(mm.evaluate_rate(0,1), 2.0)
        self.assertEqual(mm.evaluate_rate(1,2), 2.0)
        self.assertEqual(mm.evaluate_rate(2,3), 2.0)

        self.assertEqual(mm.evaluate_rate(1,2), 2.0)
        self.assertEqual(mm.evaluate_rate(2,3), 2.0)
        self.assertEqual(mm.evaluate_rate(1,0), 1.0)
        self.assertEqual(mm.evaluate_rate(2,1), 1.0)
        self.assertEqual(mm.evaluate_rate(3,2), 1.0)

        self.assertEqual(mm.evaluate_rate(0,0), 0.0)
        self.assertEqual(mm.evaluate_rate(0,2), 0.0)
        self.assertEqual(mm.evaluate_rate(0,3), 0.0)

        self.assertEqual(mm.evaluate_rate(1,1), 0.0)
        self.assertEqual(mm.evaluate_rate(1,3), 0.0)

        self.assertEqual(mm.evaluate_rate(2,0), 0.0)
        self.assertEqual(mm.evaluate_rate(2,2), 0.0)

        self.assertEqual(mm.evaluate_rate(3,0), 0.0)
        self.assertEqual(mm.evaluate_rate(3,1), 0.0)
        self.assertEqual(mm.evaluate_rate(3,3), 0.0)

    def test_state_transition(self):
        # Create Constant MarkovModel
        mm = jit(constant_model, gillstepper)

        # State 0
        self.assertEqual(mm.state(), 0)
        self.assertEqual(mm.value(), 0.0)
        self.assertEqual(mm.evaluate_rate(0), 2.0)
        self.assertEqual(mm.change_state(0), False)
       
        # State 1
        self.assertEqual(mm.state(), 1)
        self.assertEqual(mm.value(), 0.0)
        self.assertEqual(mm.evaluate_rate(0), 1.0)
        self.assertEqual(mm.evaluate_rate(1), 2.0)
        self.assertEqual(mm.change_state(1), False)
       
        # State 2
        self.assertEqual(mm.state(), 2)
        self.assertEqual(mm.value(), 0.0)
        self.assertEqual(mm.evaluate_rate(0), 1.0)
        self.assertEqual(mm.evaluate_rate(1), 2.0)
        self.assertEqual(mm.change_state(1), True)
       
        # State 3
        self.assertEqual(mm.state(), 3)
        self.assertEqual(mm.value(), 1.0)
        self.assertEqual(mm.evaluate_rate(0), 1.0)
        self.assertEqual(mm.change_state(0), True)
       
        # State 2
        self.assertEqual(mm.state(), 2)
        self.assertEqual(mm.value(), 0.0)
        self.assertEqual(mm.evaluate_rate(0), 1.0)
        self.assertEqual(mm.evaluate_rate(1), 2.0)
        self.assertEqual(mm.change_state(0), False)
       
        # State 1
        self.assertEqual(mm.state(), 1)
        self.assertEqual(mm.value(), 0.0)
        self.assertEqual(mm.evaluate_rate(0), 1.0)
        self.assertEqual(mm.evaluate_rate(1), 2.0)
        self.assertEqual(mm.change_state(0), False)
       
        # State 0
        self.assertEqual(mm.state(), 0)
        self.assertEqual(mm.value(), 0.0)
        self.assertEqual(mm.evaluate_rate(0), 2.0)

    def test_userspecified_init(self):
        mm = jit(constant_model, gillstepper)
        mm.init_state = 3
        mm.reset()
        self.assertEqual(mm.state(), 3)
        mm.fix_state = True
        self.assertEqual(mm.change_state(0), False)
        self.assertEqual(mm.state(), 3)
        
    def test_temp_setting(self):
        # Create Constant MarkovModel
        set_T(37)                         
        mm = jit(constant_model, gillstepper)

        self.assertEqual(mm.evaluate_rate(0,1), 2.0)
        self.assertEqual(mm.evaluate_rate(3,2), 1.0)
                                           
        set_T(27)
        mm.update_temp_scale()
        self.assertEqual(mm.evaluate_rate(0,1), 1.0)
        self.assertEqual(mm.evaluate_rate(3,2), 0.5)
                                           
        set_T(47)                          
        mm.update_temp_scale()             
        
        self.assertEqual(mm.evaluate_rate(0,1), 4.0)
        self.assertEqual(mm.evaluate_rate(3,2), 2.0)
                                           
        mm.set_parameter("T_ref", 47.0)    
        self.assertEqual(mm.evaluate_rate(0,1), 2.0)
        self.assertEqual(mm.evaluate_rate(3,2), 1.0)
                                           
        mm.set_parameter("T_ref", 37.0)    
        mm.set_parameter("Q10", 1.0)       
         
        self.assertEqual(mm.evaluate_rate(0,1), 2.0)
        self.assertEqual(mm.evaluate_rate(3,2), 1.0)

    def test_derivatives(self):
        mm = jit(simple_model, gillstepper)
        mm.receive("V")(10)
        mm.receive("dyadic_Ca")(0.1)
        gillstepper.set_deterministic(True, 0.9999999999)
        mm.reset()
        self.assertEqual(mm.state(), 3)
        
        mm.receive("V")(20)
        mm.receive("dyadic_Ca")(10)
        self.assertAlmostEqual(mm.substep_size(0.0), 0.0010101010101)

        gillstepper.set_max_fractional_rate_change(0.01)
        self.assertAlmostEqual(mm.substep_size(0.0), 0.00010101010101)
        rate0 = mm.evaluate_substep_rate(0, 0.0, 0.0)
        rate1 = mm.evaluate_substep_rate(0, 0.0, 0.00010101010101)
        self.assertAlmostEqual(0.01, (rate1-rate0)/rate0)
        
        mm.receive("V")(20)
        mm.receive("dyadic_Ca")(10)
        
        self.assertEqual(mm.substep_size(0.0), 1.0)

    def test_connection(self):
        mm = jit(constant_model, gillstepper)

        def receive(value):
            global received_value
            received_value = value

        class Receiver:
            def __init__(self):
                self.value = 1.0
            def receive(self, value):
                self.value = value

        receiver = Receiver()

        mm.connect(receive)
        mm.connect(receiver.receive)
        
        del receive
        
        self.assertEqual(mm.state(), 0)
        self.assertEqual(mm.value(), 0.0)
        
        mm.send()
        
        self.assertEqual(received_value, 0.0)
        self.assertEqual(receiver.value, 0.0)

        
if __name__ == "__main__":
    unittest.main()
